"""Path complex."""

from collections.abc import Hashable, Iterable, Iterator, Sequence
from typing import Any, Self

import networkx as nx
import numpy as np
import scipy as sp
from networkx.classes.reportviews import EdgeView, NodeView

from toponetx.classes.complex import Complex
from toponetx.classes.path import Path
from toponetx.classes.reportviews import PathView

__all__ = ["PathComplex"]


class PathComplex(Complex):
    """A class representing a path complex based on simple paths as proposed in [1]_.

    The original path complex is defined in [2]_.

    A path complex contains elementary p-paths that span the space of simple paths.
    Path complexes are a topological structure whose building blocks are paths, which
    are essentially different from simplicial complexes and cell complexes. If certain
    conditions are met, path complexes can generalize simplicial complexes.

    For example, a triangle with three vertices 1, 2, and 3 can be represented as a
    simplicial complex (1, 2, 3). Similarly, it can be represented as a path complex
    (1, 2, 3) whose boundary 1-paths are (1, 2), (2, 3), and (1, 3). Another example is
    a simple path (1, 2, 3). In this case, we cannot lift the simple path to a
    2-dimensional simplicial complex, as triangle does not exist. However, we can still
    represent the simple path as a path complex (1, 2, 3) whose boundary 1-paths are
    (1, 2) and (2, 3) (1-path (1,3) does not exist).

    Parameters
    ----------
    paths : nx.Graph or Iterable[Sequence[Hashable]]
        The paths in the path complex. If a graph is provided, the path complex will be
        constructed from the graph, and allowed paths are automatically computed.
    reserve_sequence_order : bool, default=False
        If True, reserve the order of the sub-sequence of nodes in the elementary
        p-path. Else, the sub-sequence of nodes in the elementary p-path will be
        reversed if the first index is larger than the last index.
    allowed_paths : Iterable[tuple[Hashable]], optional
        An iterable of allowed boundaries.
        If `None` and the input is a Graph, allowed paths are automatically computed by
        enumerating all simple paths in the graph whose length is less than or equal to
        max_rank.
        If None and the input is not a Graph, allowed paths contain the input paths,
        their truncated subsequences (sub-sequences where the first or the last index
        is omitted), and any sub-sequences of the truncated subsequences in a recursive
        manner.
    max_rank : int, default=3
        The maximal length of a path in the path complex.
    **kwargs : keyword arguments, optional
        Additional attributes to be associated with the path complex.

    Notes
    -----
    - By the definition established by [2]_, a path complex P is a non-empty collection
      of elementary p-paths such that for any sequence of vertices in a finite
      non-empty set V that belong to P, the truncated sequence of vertices (which is
      sometimes referred to as obvious boundaries) also belongs to P. The truncated
      sequences are sub-sequences of the original elementary p-path where the first or
      the last index is omitted. For instance, if we have an elementary p-path
      (1, 2, 3), the truncated sequences are (1,2) and (2,3).
    - Path complex in [1]_ is different from the path complex defined in [2]_ in the
      sense that elementary p-paths span the space of simple paths. The path complex
      originally proposed has elementary p-paths that span the space of
      boundary-invariant paths.
    - Unlike hypergraphs, combinatorial, simplicial, and cell complexes, path complexes
      take into account the order of sequences.

    References
    ----------
    .. [1] Truong and Chin.
        Generalizing Topological Graph Neural Networks with Paths.
        https://arxiv.org/abs/2308.06838.pdf
    .. [2] Grigor'yan, Lin, Muranov, and Yau.
        Homologies of path complexes and digraphs.
        https://arxiv.org/pdf/1207.2834.pdf

    Examples
    --------
    >>> PC = tnx.PathComplex([(1, 2, 3)])
    >>> PC.paths
    PathView([(1,), (2,), (3,), (1, 2), (2, 3), (1, 2, 3)])
    >>> PC.add_paths_from([(1, 2, 4), (1, 2, 5), (4, 5)])
    >>> PC.paths
    PathView([(1,), (2,), (3,), (4,), (5,), (1, 2), (2, 3), (2, 4), (2, 5), (4, 5), (1, 2, 3), (1, 2, 4), (1, 2, 5)])
    >>> G = nx.Graph()
    >>> G.add_edges_from([(1, 2), (2, 3), (1, 3)])
    >>> PC = tnx.PathComplex(G)
    >>> PC.paths
    PathView([(1,), (2,), (3,), (1, 2), (1, 3), (2, 3), (1, 3, 2), (1, 2, 3), (2, 1, 3)])
    """

    def __init__(
        self,
        paths: nx.Graph | Iterable[Sequence[Hashable] | Path] | None = None,
        reserve_sequence_order: bool = False,
        allowed_paths: Iterable[tuple[Hashable]] | None = None,
        max_rank: int = 3,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self._path_set = PathView()
        self._G = nx.Graph()
        self._reserve_sequence_order = reserve_sequence_order
        if allowed_paths is not None:
            if len(allowed_paths) > 0:
                for i in range(
                    len(allowed_paths)
                ):  # make sure list does not contain list
                    if isinstance(allowed_paths[i], list):
                        allowed_paths[i] = tuple(allowed_paths[i])
                self._allowed_paths = set(allowed_paths)
            else:
                self._allowed_paths = set()
        else:
            self._allowed_paths = set()

        if isinstance(paths, nx.Graph):
            # compute allowed_paths in order to construct boundary incidence matrix/adj matrix.
            self._G = paths
            if self._allowed_paths is not None:
                self._allowed_paths.update(
                    self.compute_allowed_paths(
                        paths,
                        reserve_sequence_order=reserve_sequence_order,
                        max_rank=max_rank,
                    )
                )

            # get feature of nodes and edges if available
            for path, data in paths.nodes(data=True):
                self.add_path(path, **data)
            for u, v, data in paths.edges(
                data=True
            ):  # so far, path complex only supports undirected graph
                if (str(u) > str(v)) and not reserve_sequence_order:
                    u, v = v, u
                self.add_path((u, v), **data)

            # add all simple paths
            self.add_paths_from(self._allowed_paths)

        elif isinstance(paths, list | tuple):
            tmp_paths = []
            for path in paths:
                if len(path) <= max_rank + 1:
                    if isinstance(paths[0], int | str | list | tuple):
                        tmp_paths.append(tuple(path))
                    else:  # path is a Path object
                        tmp_paths.append(path)
            self.add_paths_from(tmp_paths)
        elif paths is not None:
            raise TypeError(
                "Input paths must be a graph or an iterable of paths as lists or tuples."
            )

    def add_paths_from(self, paths: Iterable[Sequence[Hashable] | Path]) -> None:
        """Add elementary paths from an iterable of elementary paths.

        An elementary p-path is a sequence of nodes $(n_1, ..., n_p)$ where $p$ is the length of the sequence. In a path complex,
        for every elementary p-path in the path complex, their truncated sequences $(n_2, ..., n_p)$ and $(n1, ..., n_{p-1})$ are also in the path complex.

        Parameters
        ----------
        paths : Iterable[Sequence[Hashable] or Path]
            An iterable of paths as lists, tuples, or Path objects.

        Examples
        --------
        >>> PC = tnx.PathComplex([(1, 2, 3)])
        >>> PC.paths
        PathView([(1,), (2,), (3,), (1, 2), (2, 3), (1, 2, 3)])
        >>> PC.add_paths_from([(1, 2, 4), (1, 2, 5), (4, 5)])
        >>> PC.paths
        PathView([(1,), (2,), (3,), (4,), (5,), (1, 2), (2, 3), (2, 4), (2, 5), (4, 5), (1, 2, 3), (1, 2, 4), (1, 2, 5)])
        """
        paths_clone = paths.copy()
        for p in paths_clone:
            self.add_path(p)

    def add_path(self, path: Hashable | Sequence[Hashable] | Path, **attr) -> None:
        """Add an elementary path to the path complex.

        An elementary p-path is a sequence of nodes $(n_1, ..., n_p)$ where p is the length of the sequence. In a path complex,
        for every elementary p-path in the path complex, their truncated sequences $(n_2, ..., n_p)$ and $(n_1, ..., n_{p-1})$ are also in the path complex.

        This method automatically initializes any obvious sub-paths (sub-paths where the first or last index is omitted) of the elementary path if not available.
        In order to add non-obvious sub-paths, manually add the sub-paths.

        Parameters
        ----------
        path : Hashable or Sequence[Hashable] or Path
            A Hashable or Sequence[Hashable] or Path representing a path in a path complex.
        **attr : keyword arguments, optional
            Additional keyword arguments to be passed to the method.

        Examples
        --------
        >>> PC = tnx.PathComplex([(1, 2, 3)])
        >>> PC.paths
        PathView([(1,), (2,), (3,), (1, 2), (2, 3), (1, 2, 3)])
        >>> PC.add_path((1, 2, 4))
        >>> PC.paths
        PathView([(1,), (2,), (3,), (4,), (1, 2), (2, 3), (2, 4), (1, 2, 3), (1, 2, 4)])
        """
        new_paths = set()
        if isinstance(path, int | str):
            path = [path]
        if isinstance(path, list | tuple | Path):
            if not isinstance(path, Path):  # path is a list or tuple
                path_ = tuple(path)
                if len(path) != len(set(path)):
                    raise ValueError(
                        "An elementary p-path cannot contain duplicate nodes."
                    )
                if (
                    len(path_) > 1
                    and str(path_[0]) > str(path_[-1])
                    and not self._reserve_sequence_order
                ):
                    raise ValueError(
                        f"An elementary p-path must have the first index smaller than the last index, got {path}."
                    )
            else:  # path is a Path object
                path_ = path.elements
            self._update_faces_dict_length(
                path_
            )  # add dict corresponding to the path dimension

            if (
                self._path_set.max_dim < len(path_) - 1
            ):  # update max dimension for PathView()
                self._path_set.max_dim = len(path_) - 1

            if (
                path_ in self._path_set.faces_dict[len(path_) - 1]
            ):  # path is already in the complex, just update the attributes if needed
                self._update_attributes(path, **attr)
                return

            # update sub-paths
            for length in range(len(path_), 0, -1):
                for i in range(len(path_) - length + 1):
                    sub_path = path_[i : i + length]
                    if not self._reserve_sequence_order and str(sub_path[0]) > str(
                        sub_path[-1]
                    ):
                        sub_path = sub_path[::-1]
                    sub_path = tuple(sub_path)

                    # add to _G
                    if len(sub_path) == 1:
                        self._G.add_node(sub_path[0])
                    elif len(sub_path) == 2:
                        self._G.add_edge(sub_path[0], sub_path[1])

                    # expand _path_set if necessary. keep track of newly added paths to expend _allowed_paths
                    new_path = self._update_faces_dict_entry(sub_path)
                    if new_path is not None:
                        new_paths.add(new_path)
            # update allowed paths
            if len(new_paths) > 0:
                self._allowed_paths.update(new_paths)

            self._update_attributes(path, **attr)

    @property
    def dim(self) -> int:
        """Dimension.

        Returns
        -------
        int
            This is the highest dimension of any elementary p-path in the complex.
        """
        return self._path_set.max_dim

    @property
    def nodes(self) -> NodeView:
        """Nodes.

        Returns
        -------
        NodeView
            A view of all nodes in the path complex.
        """
        return self._G.nodes

    @property
    def edges(self) -> EdgeView:
        """Edges.

        Returns
        -------
        EdgeView
            A view of all edges in the path complex.
        """
        return self._G.edges

    @property
    def paths(self) -> PathView:
        """Set of all elementary p-paths.

        Returns
        -------
        PathView
            A view of all elementary p-paths in the path complex.
        """
        return self._path_set

    @property
    def shape(self) -> tuple[int, ...]:
        """Shape of path complex.

        (number of elementary p-paths for each dimension d, for d in range(0,dim(Pc)))

        Returns
        -------
        tuple[int, ...]
            Returns a tuple representing the shape of the path complex.
        """
        return self._path_set.shape

    def clone(self) -> Self:
        """Return a copy of the path complex.

        The clone method by default returns an independent shallow copy of the path
        complex. Use Python's `copy.deepcopy` for new containers.

        Returns
        -------
        PathComplex
            Returns a copy of the PathComplex.
        """
        return self.__class__(list(self.paths), rank=self.dim)

    def skeleton(self, rank: int) -> list[tuple[Hashable]]:
        """Compute skeleton.

        Parameters
        ----------
        rank : int
            The rank to be used to generate the skeleton of elementary p-paths.

        Returns
        -------
        set[tuple[Hashable]]
            Set of elementary p-paths of dimension specified by `rank`.
        """
        if len(self._path_set.faces_dict) > rank >= 0:
            tmp = (path for path in self._path_set.faces_dict[rank])
            return sorted(
                tmp, key=lambda x: tuple(map(str, x))
            )  # lexicographic comparison
        if rank < 0:
            raise ValueError(f"input must be a positive integer, got {rank}")
        raise ValueError(f"input {rank} exceeds max dim")

    def add_node(self, node: Hashable | Path, **attr) -> None:
        """Add node to the path complex.

        Parameters
        ----------
        node : Hashable or Path
            A Hashable or singleton Path representing a node in a path complex.
        **attr : keyword arguments
            Additional attributes to be provided as keyword arguments.
        """
        if not isinstance(node, Hashable):
            raise TypeError(f"Input node must be Hashable, got {type(node)} instead.")

        if isinstance(node, Path):
            if len(node) != 1:
                raise ValueError(
                    f"Input node must be a singleton Path, got {node} instead."
                )
            self.add_path(node, **attr)
        else:
            self.add_path([node], **attr)

    def remove_nodes(self, node_set: Iterable[Hashable]) -> None:
        """Remove nodes from the path complex.

        Parameters
        ----------
        node_set : Iterable[Hashable]
            An iterable of nodes to be removed.
        """
        removed_paths = set()
        for path in self.paths:  # iterate over all paths
            if any(
                node in path for node in node_set
            ):  # if any node in node_set is in the path, remove the path
                removed_paths.add(path)

        for path in removed_paths:
            self._remove_path(path)

        self._G.remove_nodes_from(node_set)

    def incidence_matrix(
        self,
        rank: int,
        signed: bool = True,
        weight: str | None = None,
        index: bool = False,
    ) -> tuple[dict, dict, sp.sparse.lil_matrix] | sp.sparse.lil_matrix:
        """Compute incidence matrix of the path complex.

        Parameters
        ----------
        rank : int
            The dimension of the incidence matrix.
        signed : bool, default=True
            If True, return signed incidence matrix. Else, return absolute incidence matrix.
        weight : str, optional
            If not given, all nonzero entries are 1.
        index : bool, default=False
            If True, return incidence matrix with indices. Else, return incidence matrix without indices.

        Returns
        -------
        tuple[dict, dict, sp.sparse.lil_matrix] | sp.sparse.lil_matrix
            If `index` is True, return a tuple of (idx_p_minus_1, idx_p, incidence_matrix) otherwise
            if `index` is False, return incidence_matrix.
        """
        if rank < 0:
            raise ValueError(f"input dimension d must be positive integer, got {rank}")
        if rank > self.dim:
            raise ValueError(
                f"input dimension cannot be larger than the dimension of the complex, got {rank}"
            )
        if rank == 0:
            boundary = sp.sparse.lil_matrix((0, len(self.nodes)))
            if index:
                node_index = {
                    (node,): i for i, node in enumerate(sorted(self.nodes, key=str))
                }
                return {}, node_index, abs(boundary.tocoo())
            return abs(boundary.tocoo())

        idx_p_minus_1, idx_p, values = [], [], []
        path_minus_1_dict = {
            path: i for i, path in enumerate(self.skeleton(rank - 1))
        }  # path2idx dict
        path_dict = {
            path: i for i, path in enumerate(self.skeleton(rank))
        }  # path2idx dict
        for path, idx_path in path_dict.items():
            for i, _ in enumerate(path):
                boundary_path = path[0:i] + path[(i + 1) :]
                if not self._reserve_sequence_order and str(boundary_path[0]) > str(
                    boundary_path[-1]
                ):
                    boundary_path = boundary_path[::-1]
                boundary_path = tuple(boundary_path)
                if boundary_path in self._allowed_paths:
                    idx_p_minus_1.append(path_minus_1_dict[boundary_path])
                    idx_p.append(idx_path)
                    values.append((-1) ** i)
        boundary = sp.sparse.coo_matrix(
            (values, (idx_p_minus_1, idx_p)),
            dtype=np.float32,
            shape=(
                len(path_minus_1_dict),
                len(path_dict),
            ),
        )

        if index:
            if signed:
                return (
                    path_minus_1_dict,
                    path_dict,
                    boundary,
                )

            return (
                path_minus_1_dict,
                path_dict,
                abs(boundary),
            )

        if signed:
            return boundary
        return abs(boundary)

    def up_laplacian_matrix(
        self,
        rank: int,
        signed: bool = True,
        weight: str | None = None,
        index: bool = False,
    ) -> tuple[dict, sp.sparse.lil_matrix] | sp.sparse.lil_matrix:
        """Compute up laplacian matrix of the path complex.

        Parameters
        ----------
        rank : int
            The dimension of the up laplacian matrix.
        signed : bool, default=True
            If True, return signed up laplacian matrix. Else, return absolute up laplacian matrix.
        weight : str, optional
            If not given, all nonzero entries are 1.
        index : bool, default=False
            If True, return up laplacian matrix with indices. Else, return up laplacian matrix without indices.

        Returns
        -------
        tuple[dict, sp.sparse.lil_matrix] | sp.sparse.lil_matrix
            If `index` is True, return a tuple of (idx_p, up_laplacian_matrix) otherwise
            if `index` is False, return up_laplacian_matrix.
        """
        if weight is not None:
            raise ValueError("Weighted Laplacian is not supported in this version.")

        if 0 <= rank < self.dim:
            row, _, B_next = self.incidence_matrix(rank + 1, weight=weight, index=True)
            L_up = B_next @ B_next.transpose()
        else:
            raise ValueError(
                f"Rank should larger than 0 and <= {self.dim - 1} (maximal dimension-1), got {rank}."
            )
        if not signed:
            L_up = abs(L_up)

        if index:
            return row, L_up.tolil()
        return L_up.tolil()

    def down_laplacian_matrix(
        self,
        rank: int,
        signed: bool = True,
        weight: str | None = None,
        index: bool = False,
    ) -> tuple[dict, sp.sparse.lil_matrix] | sp.sparse.lil_matrix:
        """Compute down laplacian matrix of the path complex.

        Parameters
        ----------
        rank : int
            The dimension of the down laplacian matrix.
        signed : bool, default=True
            If True, return signed down laplacian matrix. Else, return absolute down laplacian matrix.
        weight : str, optional
            If not given, all nonzero entries are 1.
        index : bool, default=False
            If True, return down laplacian matrix with indices. Else, return down laplacian matrix without indices.

        Returns
        -------
        tuple[dict, sp.sparse.lil_matrix] | sp.sparse.lil_matrix
            If `index` is True, return a tuple of (idx_p, down_laplacian_matrix) otherwise
            if `index` is False, return down_laplacian_matrix.
        """
        if weight is not None:
            raise ValueError("Weighted Laplacian is not supported in this version.")

        if self.dim >= rank > 0:
            row, _, B = self.incidence_matrix(rank, weight=weight, index=True)
            L_down = B.transpose() @ B
        else:
            raise ValueError(
                f"Rank should be larger than 1 and <= {self.dim} (maximal dimension), got {rank}."
            )
        if not signed:
            L_down = abs(L_down)
        if index:
            return row, L_down.tolil()
        return L_down.tolil()

    def hodge_laplacian_matrix(
        self,
        rank: int,
        signed: bool = True,
        weight: str | None = None,
        index: bool = False,
    ) -> tuple[dict, sp.sparse.lil_matrix] | sp.sparse.lil_matrix:
        """Compute Hodge Laplacian matrix of the path complex.

        Parameters
        ----------
        rank : int
            The dimension of the Hodge Laplacian matrix.
        signed : bool, default=False
            If True, return signed Hodge Laplacian matrix. Else, return absolute Hodge Laplacian matrix.
        weight : str, optional
            If not given, all nonzero entries are 1.
        index : bool, default=False
            If True, return Hodge Laplacian matrix with indices. Else, return Hodge Laplacian matrix without indices.

        Returns
        -------
        tuple[dict, sp.sparse.lil_matrix] | sp.sparse.lil_matrix
            When index is True, return a tuple of (idx_p, Laplacian) else return Laplacian.
        """
        if rank == 0:
            row, column, B_next = self.incidence_matrix(
                rank + 1, weight=weight, index=True
            )
            L_hodge = B_next @ B_next.transpose()
            if not signed:
                L_hodge = abs(L_hodge)
            if index:
                return row, L_hodge
            return L_hodge
        if rank < self.dim:
            row, column, B_next = self.incidence_matrix(
                rank + 1, weight=weight, index=True
            )
            row, column, B = self.incidence_matrix(rank, weight=weight, index=True)
            L_hodge = B_next @ B_next.transpose() + B.transpose() @ B
            if not signed:
                L_hodge = abs(L_hodge)
            if index:
                return column, L_hodge
            return L_hodge
        if rank == self.dim:
            row, column, B = self.incidence_matrix(rank, weight=weight, index=True)
            L_hodge = B.transpose() @ B
            if not signed:
                L_hodge = abs(L_hodge)
            if index:
                return column, L_hodge
            return L_hodge

        raise ValueError(
            f"Rank should be larger than 0 and <= {self.dim} (maximal dimension simplices), got {rank}"
        )

    def adjacency_matrix(
        self,
        rank: int,
        signed: bool = False,
        weight: str | None = None,
        index: bool = False,
    ) -> tuple[dict, sp.sparse.lil_matrix] | sp.sparse.lil_matrix:
        """Compute adjacency matrix of the path complex.

        Parameters
        ----------
        rank : int
            The dimension of the adjacency matrix.
        signed : bool, default=False
            If True, return signed adjacency matrix. Else, return absolute adjacency matrix.
        weight : str, default=None, optional
            The weight to be used for the adjacency matrix.
        index : bool, default=False
            If True, return adjacency matrix with indices. Else, return adjacency matrix without indices.

        Returns
        -------
        tuple[dict, sp.sparse.lil_matrix] | sp.sparse.lil_matrix
            If `index` is True, return a tuple of (idx_p, adjacency_matrix) else return adjacency_matrix.
        """
        ind, L_up = self.up_laplacian_matrix(rank, signed=signed, index=True)
        L_up.setdiag(0)

        if not signed:
            L_up = abs(L_up)
        if index:
            return ind, L_up
        return L_up

    def coadjacency_matrix(
        self,
        rank: int,
        signed: bool = False,
        weight: str | None = None,
        index: bool = False,
    ) -> tuple[dict, sp.sparse.lil_matrix] | sp.sparse.lil_matrix:
        """Compute coadjacency matrix of the path complex.

        Parameters
        ----------
        rank : int
            The dimension of the coadjacency matrix.
        signed : bool, default=False
            If True, return signed coadjacency matrix. Else, return absolute coadjacency matrix.
        weight : str, default=None, optional
            The weight to be used for the adjacency matrix.
        index : bool, default=False
            If True, return coadjacency matrix with indices. Else, return coadjacency matrix without indices.

        Returns
        -------
        tuple[dict, sp.sparse.lil_matrix] | sp.sparse.lil_matrix
            If `index` is True, return a tuple of (idx_p, coadjacency_matrix) else return coadjacency_matrix.
        """
        ind, L_down = self.down_laplacian_matrix(rank, signed=signed, index=True)
        L_down.setdiag(0)
        if not signed:
            L_down = abs(L_down)
        if index:
            return ind, L_down
        return L_down

    def restrict_to_nodes(self, node_set: Iterable[Hashable]):
        """Return a new path complex restricted to a subset of nodes.

        Parameters
        ----------
        node_set : Iterable[Hashable]
            The nodes to be used for the new restricted path complex.

        Returns
        -------
        PathComplex
            The new path complex restricted to the provided subset of nodes.

        Examples
        --------
        >>> PC = tnx.PathComplex(
        ...     [[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3], [0, 1, 2], [0, 1, 3]]
        ... )
        >>> PC = PC.restrict_to_nodes([0, 1, 3])
        >>> PC.paths
        PathView([(0,), (1,), (3,), (0, 1), (1, 3), (0, 1, 3)])
        """
        if len(node_set) == 0:
            raise ValueError("Input node_set cannot be empty.")
        node_set = set(node_set)
        new_paths = [
            path for path in self.paths if all(node in node_set for node in path)
        ]
        return PathComplex(new_paths)

    def restrict_to_paths(self, path_set: Iterable[Sequence[Hashable]]):
        """Return a new path complex restricted to a subset of paths.

        Parameters
        ----------
        path_set : Iterable[Sequence[Hashable]]
            The paths to be used for the new restricted path complex.

        Returns
        -------
        PathComplex
            The new path complex restricted to the provided subset of paths.

        Examples
        --------
        >>> PC = tnx.PathComplex(
        ...     [[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3], [0, 1, 2], [0, 1, 3]]
        ... )
        >>> PC = PC.restrict_to_paths([[1, 2, 3]])
        >>> PC.paths
        PathView([(1,), (2,), (3,), (1, 2), (2, 3), (1, 2, 3)])
        """
        if len(path_set) == 0:
            raise ValueError("Input path_set cannot be empty.")
        new_path_set = [tuple(path) for path in path_set]
        new_path_set = set(new_path_set)
        new_paths = [path for path in self.paths if path in new_path_set]
        return PathComplex(new_paths)

    def get_node_attributes(self, name: str) -> dict[tuple[Hashable], Any]:
        """Get node attributes from a path complex.

        Parameters
        ----------
        name : str
            The name of the node attribute.

        Returns
        -------
        dict[tuple[Hashable], Any]
            A dictionary of node and its associated value for the given attribute name.

        Examples
        --------
        >>> PC = tnx.PathComplex()
        >>> PC.add_node(0)
        >>> PC.add_node(1, heat=55)
        >>> PC.add_node(2, heat=66)
        >>> PC.add_node(3, color="red")
        >>> PC.add_node(2, color="blue")
        >>> PC.add_paths_from([[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3]])
        >>> PC.get_node_attributes("heat")
        {(1,): 55, (2,): 66}
        >>> PC.get_node_attributes("color")
        {(2,): 'blue', (3,): 'red'}
        """
        return {tuple(n): self[n][name] for n in self.skeleton(0) if name in self[n]}

    def set_node_attributes(
        self,
        values: dict[Sequence[Hashable], Hashable | dict[Hashable, Any]],
        name: str | None = None,
    ) -> None:
        """Set node attributes for a path complex.

        Parameters
        ----------
        values : dict[Sequence[Hashable], Hashable | dict[Hashable, Any]]
            A dictionary of node and its associated attribute values.
        name : str
            The name of the node attribute.

        Examples
        --------
        >>> PC = tnx.PathComplex()
        >>> PC.add_paths_from([[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3]])
        >>> PC.set_node_attributes(
        ...     {
        ...         (1,): {"heat": 55, "color": "red"},
        ...         (2,): {"heat": 66, "color": "blue"},
        ...     }
        ... )
        >>> PC[1]
        {'heat': 55, 'color': 'red'}
        >>> PC[2]
        {'heat': 66, 'color': 'blue'}
        >>> PC.set_node_attributes({(1,): 58, (2,): 60}, name="heat")
        >>> PC[1]
        {'heat': 58, 'color': 'red'}
        >>> PC[2]
        {'heat': 60, 'color': 'blue'}
        """
        if name is None:
            for node, value in values.items():
                if isinstance(value, dict):
                    for k, v in value.items():
                        self[node][k] = v
                        if isinstance(node, tuple | list):
                            if len(node) == 1:
                                self._G.nodes[node[0]][k] = v
                            else:
                                raise ValueError("Input node must be a singleton.")
                        elif isinstance(node, int | str):
                            self._G.nodes[node][k] = v
                else:
                    raise TypeError(
                        "Input values must be a dictionary of node and its associated attribute values."
                    )
        else:
            for node, value in values.items():
                self[node][name] = value
                if isinstance(node, tuple | list):
                    if len(node) == 1:
                        self._G.nodes[node[0]][name] = value
                    else:
                        raise ValueError("Input node must be a singleton.")
                elif isinstance(node, int | str):
                    self._G.nodes[node][name] = value

    def get_edge_attributes(self, name: str) -> dict[tuple[Hashable], Any]:
        """Get edge attributes from a path complex.

        Parameters
        ----------
        name : str
            The name of the edge attribute.

        Returns
        -------
        dict[tuple[Hashable], Any]
            A dictionary of edge and its associated value for the given attribute name.

        Examples
        --------
        >>> PC = tnx.PathComplex()
        >>> PC.add_paths_from([[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3]])
        >>> PC.add_path([0, 1], weight=32)
        >>> PC.add_path([1, 2], weight=98)
        >>> PC.add_path([1, 3], color="red")
        >>> PC.add_path([2, 3], color="blue")
        >>> PC.get_edge_attributes("weight")
        {(0, 1): 32, (1, 2): 98}
        >>> PC.get_edge_attributes("color")
        {(1, 3): 'red', (2, 3): 'blue'}
        """
        return {tuple(e): self[e][name] for e in self.skeleton(1) if name in self[e]}

    def set_edge_attributes(
        self,
        values: dict[Sequence[Hashable], Hashable | dict[Hashable, Any]],
        name: str | None = None,
    ) -> None:
        """Set edge attributes for a path complex.

        Parameters
        ----------
        values : dict[Sequence[Hashable], Hashable | dict[Hashable, Any]]
            A dictionary of edge and its associated attribute values.
        name : str
            The name of the edge attribute.

        Examples
        --------
        >>> PC = tnx.PathComplex()
        >>> PC.add_paths_from([[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3]])
        >>> PC.add_path([0, 1], weight=32)
        >>> PC.add_path([1, 2], weight=98)
        >>> PC.add_path([1, 3], color="red")
        >>> PC.add_path([2, 3], color="blue")
        >>> PC.get_edge_attributes("weight")
        {(0, 1): 32, (1, 2): 98}
        >>> PC.set_edge_attributes({(0, 1): 33}, name="weight")
        >>> PC.get_edge_attributes("weight")
        {(0, 1): 33, (1, 2): 98}
        >>> PC.set_edge_attributes(
        ...     {
        ...         (1, 3): {"weight": 55, "color": "yellow"},
        ...         (2, 3): {"weight": 66, "color": "blue"},
        ...     }
        ... )
        >>> PC[1, 3]
        {'color': 'yellow', 'weight': 55}
        >>> PC[2, 3]
        {'color': 'blue', 'weight': 66}
        """
        if name is None:
            for edge, value in values.items():
                if len(edge) != 2:
                    raise ValueError("Input edge must be a pair.")
                if isinstance(value, dict):
                    for k, v in value.items():
                        self[edge][k] = v
                        if isinstance(edge, tuple | list):
                            self._G.edges[edge[0], edge[1]][k] = v
                else:
                    raise TypeError(
                        "Input values must be a dictionary of edge and its associated attribute values."
                    )
        else:
            for edge, value in values.items():
                if len(edge) != 2:
                    raise ValueError("Input edge must be a pair.")
                self[edge][name] = value
                if isinstance(edge, tuple | list):
                    self._G.edges[edge[0], edge[1]][name] = value

    def get_path_attributes(self, name: str) -> dict[tuple[Hashable], Any]:
        """Get path attributes from a path complex.

        Parameters
        ----------
        name : str
            The name of the path attribute.

        Returns
        -------
        dict[tuple[Hashable], Any]
            A dictionary of path and its associated value for the given attribute name.

        Examples
        --------
        >>> PC = tnx.PathComplex()
        >>> PC.add_paths_from([[0, 1]])
        >>> PC.add_path([0, 1, 2], weight=43)
        >>> PC.add_path([0, 1, 3], weight=98)
        >>> PC.add_path([1, 2, 3], color="red")
        >>> PC.add_path([1, 3, 2], color="blue")
        >>> PC.add_path([2, 1, 3], color="green")
        >>> PC.get_path_attributes("weight")
        {(0, 1, 2): 43, (0, 1, 3): 98}
        >>> PC.get_path_attributes("color")
        {(1, 2, 3): 'red', (1, 3, 2): 'blue', (2, 1, 3): 'green'}
        """
        return {
            tuple(p): self[p][name]
            for dim in range(self.dim + 1)
            for p in self.skeleton(dim)
            if name in self[p]
        }

    def set_path_attributes(
        self,
        values: dict[Sequence[Hashable], Hashable | dict[Hashable, Any]],
        name: str | None = None,
    ) -> None:
        """Set path attributes for a path complex.

        Parameters
        ----------
        values : dict[Sequence[Hashable], Hashable | dict[Hashable, Any]]
            A dictionary of path and its associated attribute values.
        name : str
            The name of the path attribute.

        Examples
        --------
        >>> PC = tnx.PathComplex(
        ...     [[0, 1], [0, 1, 2], [0, 1, 3], [1, 2, 3], [1, 3, 2], [2, 1, 3]]
        ... )
        >>> PC.set_path_attributes(
        ...     {
        ...         (0, 1, 2): {"weight": 43, "color": "red"},
        ...         (0, 1, 3): {"weight": 98, "color": "blue"},
        ...     }
        ... )
        >>> PC.get_path_attributes("weight")
        {(0, 1, 2): 43, (0, 1, 3): 98}
        >>> PC.get_path_attributes("color")
        {(0, 1, 2): 'red', (0, 1, 3): 'blue'}
        >>> PC.set_path_attributes({0: 55}, "weight")
        >>> PC.get_path_attributes("weight")
        {(0,): 55, (0, 1, 2): 43, (0, 1, 3): 98}
        >>> PC.set_path_attributes({(0, 1): "red"}, "color")
        >>> PC.get_path_attributes("color")
        {(0, 1): 'red', (0, 1, 2): 'red', (0, 1, 3): 'blue'}
        """
        if name is None:
            for path, value in values.items():
                if isinstance(value, dict):
                    for k, v in value.items():
                        if len(path) == 1:
                            self.set_node_attributes({path: {k: v}})
                        elif len(path) == 2:
                            self.set_edge_attributes({path: {k: v}})
                        else:
                            self[path][k] = v
                else:
                    raise TypeError(
                        "Input values must be a dictionary of path and its associated attribute values."
                    )
        else:
            for path, value in values.items():
                if isinstance(path, int | str) or len(path) == 1:
                    self.set_node_attributes({path: {name: value}})
                elif len(path) == 2:
                    self.set_edge_attributes({path: {name: value}})
                else:
                    self[path][name] = value

    def _remove_path(self, path: tuple[Hashable]) -> None:
        """Remove path from the Path Complex.

        Parameters
        ----------
        path : tuple[Hashable]
            The path to be removed.
        """
        del self._path_set.faces_dict[len(path) - 1][path]
        self._allowed_paths.remove(path)
        if (
            len(self._path_set.faces_dict[len(path) - 1]) == 0
            and self._path_set.max_dim == len(path) - 1
        ):  # update max dimension for PathView() if highest dimension is empty
            self._path_set.max_dim -= 1

    def _update_faces_dict_length(self, path: tuple[Hashable]) -> None:
        """Update length of dict faces in the Path Complex.

        Parameters
        ----------
        path : tuple[Hashable]
            The path to be used for the update.
        """
        if len(path) > len(self._path_set.faces_dict):
            diff = len(path) - len(self._path_set.faces_dict)
            for _ in range(diff):
                self._path_set.faces_dict.append({})

    def _update_faces_dict_entry(self, path: tuple[Hashable]):
        """Update the dict entry for faces.

        Parameters
        ----------
        path : tuple[Hashable]
            The path to update faces dict entry for.

        Returns
        -------
        PathComlex | None
            The return value.
        """
        dim = len(path) - 1
        if path not in self._path_set.faces_dict[dim]:  # Not in faces_dict
            self._path_set.faces_dict[dim][path] = {}
            return path
        return None

    def _update_attributes(self, path, **attr):
        """Update the attributes of path.

        Parameters
        ----------
        path : tuple[Hashable]
            The path to update attributes for.
        **attr : keyword arguments
            The attributes to update.
        """
        if isinstance(path, Path):
            path_ = path.elements
            attr = path._attributes | attr
        else:
            path_ = tuple(path)

        self._path_set.faces_dict[len(path_) - 1][path_].update(attr)
        if len(path_) == 1:
            self._G.add_node(path_[0], **attr)
        elif len(path_) == 2:
            self._G.add_edge(path_[0], path_[1], **attr)

    def __contains__(self, atom: Any) -> bool:
        """Check if an atom is in the path complex.

        Parameters
        ----------
        atom : Any
            The atom to check for existence in the path complex.

        Returns
        -------
        bool
            `True` if the given atom is in the path complex, `False` otherwise.
        """
        return atom in self._path_set

    def __getitem__(self, atom: Any) -> dict[Hashable, Any]:
        """Get the elementary p-path.

        Parameters
        ----------
        atom : Any
            An elementary p-path or a node in the path complex.

        Returns
        -------
        PathView
            The pathview based on the item provided.

        Raises
        ------
        KeyError
            If the path does not exist in this path complex.
        """
        return self._path_set[atom]

    def __iter__(self) -> Iterator:
        """Iterate over all faces of the path complex.

        Returns
        -------
        Iterator
            Iterator for iterating over all faces of the path complex.
        """
        return iter(self._path_set)

    def __len__(self) -> int:
        """Return the number of elementary p-paths in the path complex.

        Returns
        -------
        int
            The number of elementary p-paths in the path complex.
        """
        return len(list(self.__iter__()))

    def __str__(self) -> str:
        """Return detailed string representation.

        Returns
        -------
        str
            The __str__ representation of the Path Complex.
        """
        return f"Path Complex with shape {self.shape} and dimension {self.dim}"

    def __repr__(self) -> str:
        """Return string representation.

        Returns
        -------
        str
            The __repr__ representation of the Path Complex.
        """
        return "PathComplex()"

    @staticmethod
    def compute_allowed_paths(
        graph: nx.Graph, reserve_sequence_order: bool = False, max_rank: int = 3
    ) -> set[list | tuple]:
        """Compute allowed paths from a graph.

        Allowed paths are automatically computed by enumerating all simple paths in the graph whose length is less than
        or equal to max_rank. Allowed paths allow us to restrict the boundary of an elementary p-path to only sequences that exist in the graph.

        Parameters
        ----------
        graph : nx.Graph
            A graph.
        reserve_sequence_order : bool, default=False
            If True, reserve the order of the sub-sequence of nodes in the elmentary p-path.
            Else, the sub-sequence of nodes in the elementary p-path will be reversed if the first index is larger than the last index.
        max_rank : int, default=3
            The maximal length of a path in the path complex.

        Returns
        -------
        set[list | tuple]
            A set of allowed paths.

        Examples
        --------
        >>> G = nx.Graph()
        >>> G.add_edges_from([(1, 2), (2, 3), (1, 3), (0, 1)])
        >>> allowed_paths = tnx.PathComplex.compute_allowed_paths(G, max_rank=2)
        >>> allowed_paths
        {(0, 1), (1, 3), (1, 2), (2,), (1, 3, 2), (0, 1, 2), (0, 1, 3), (1, 2, 3), (2, 1, 3), (2, 3), (1,), (0,), (3,)}
        """
        allowed_paths = []
        all_nodes_list = [(node,) for node in sorted(graph.nodes, key=str)]
        all_edges_list = []
        for edge in graph.edges:
            if not reserve_sequence_order and str(edge[0]) > str(edge[1]):
                edge = edge[::-1]
            all_edges_list.append(edge)
        allowed_paths.extend(all_nodes_list)
        allowed_paths.extend(all_edges_list)

        node_ls = list(graph.nodes)
        for src_idx in range(len(node_ls)):
            for tgt_idx in range(src_idx + 1, len(node_ls)):
                all_simple_paths = list(
                    nx.all_simple_paths(
                        graph,
                        source=node_ls[src_idx],
                        target=node_ls[tgt_idx],
                        cutoff=max_rank,
                    )
                )

                for i in range(len(all_simple_paths)):
                    path = all_simple_paths[i]
                    if not reserve_sequence_order:
                        all_simple_paths[i] = (
                            path[::-1] if str(path[0]) > str(path[-1]) else path
                        )
                    all_simple_paths[i] = tuple(all_simple_paths[i])

                if len(all_simple_paths) > 0:
                    allowed_paths.extend(all_simple_paths)
        return set(allowed_paths)



"""Initialize the library with modules and other content."""

# From the following modules, we do not import all symbols, but only the module itself.
import toponetx.datasets as datasets

# From the following modules, we import all symbols.
from toponetx.algorithms import *
from toponetx.classes import *
from toponetx.exception import *
from toponetx.generators import *
from toponetx.readwrite import *
from toponetx.transform import *
from toponetx.utils import *

__all__ = ["datasets"]



"""Load Austin R. Benson's datasets.

This module provides functions to load the datasets by Austin R. Benson, published on
his website:

    https://www.cs.cornell.edu/~arb/data/

A dataset is stored in a folder with the following mandatory files:

- ``{name}-nverts.txt``: Line ``i`` contains the number of vertices in the ``i``-th simplex.

- ``{name}-simplices.txt``: Each line contains a vertex ID. Line 1 corresponds to the
  first vertex of the first simplex, line 2 to the second vertex of the first simplex,
  and so on.

Optionally, additional files can be present and are consumed by the functions in this
module:

- ``{name}-times.txt``: Line ``i`` contains the timestamp of the ``i``-th simplex.
"""

from pathlib import Path

from more_itertools import spy

from toponetx.classes.simplex import Simplex

__all__ = ["load_benson_hyperedges", "load_benson_simplices"]


def _validate_folder(folder: Path) -> str:
    """Validate that the given folder contains a valid Benson dataset.

    Parameters
    ----------
    folder : Path
        Path to the folder containing the dataset.

    Returns
    -------
    str
        The name of the dataset (the folder name).

    Raises
    ------
    ValueError
        If the folder does not exist or is not in the expected format.
    """
    if not folder.exists() or not folder.is_dir():
        raise ValueError(f"Folder `{folder}` does not exist.")

    name = folder.name

    mandatory_files = [f"{name}-nverts.txt", f"{name}-simplices.txt"]
    if not all((folder / file).exists() for file in mandatory_files):
        raise ValueError(f"Folder `{folder}` is not in an expected format.")

    return name


def load_benson_hyperedges(folder: Path | str) -> tuple[list[Simplex], list[Simplex]]:
    """Load hyperedge data from the Benson dataset format.

    Parameters
    ----------
    folder : Path | str
        Path to the folder containing the dataset.

    Returns
    -------
    nodes : list[Simplex]
        List of nodes in the dataset, each represented as a `Simplex` with a single
        vertex and associated name and label.
    simplices : list[Simplex]
        List of hyperedges in the dataset.

    Raises
    ------
    ValueError
        If the folder does not exist or is not in the expected format.
    """
    if not isinstance(folder, Path):
        folder = Path(folder)

    if not folder.exists() or not folder.is_dir():
        raise ValueError(f"Folder `{folder}` does not exist.")

    name = folder.name

    if not all(
        (folder / file).exists()
        for file in [f"label-names-{name}.txt", f"node-labels-{name}.txt"]
    ):
        raise ValueError(f"Folder `{folder}` is not in an expected format.")

    with (folder / f"label-names-{name}.txt").open() as file:
        label_map = {index: line.strip() for index, line in enumerate(file, start=1)}
    with (folder / f"node-labels-{name}.txt").open() as file:
        # Peek at the first 10 lines to check for commas (multilabel) or not (single label)
        first_lines, file_iter = spy(file, 10)
        is_multilabel = any("," in line for line in first_lines)

        if is_multilabel:
            node_labels = [
                [label_map[int(idx)] for idx in line.strip().split(",")]
                for line in file_iter
            ]
        else:
            node_labels = [label_map[int(line.strip())] for line in file_iter]

    nodes = [
        Simplex([vertex], label=label)
        for vertex, label in enumerate(node_labels, start=1)
    ]

    if (folder / f"node-names-{name}.txt").exists():
        with (folder / f"node-names-{name}.txt").open() as file:
            for simplex, node_name in zip(nodes, file, strict=True):
                simplex["name"] = node_name.strip()

    with (folder / f"hyperedges-{name}.txt").open() as file:
        simplices = [
            Simplex([int(vertex) for vertex in line.split(",")]) for line in file
        ]

    return nodes, simplices


def load_benson_simplices(folder: Path | str) -> list[Simplex]:
    """Load simplicial complex data from the Benson dataset format.

    If the dataset is temporal (indicated by the presence of a ``{name}-times.txt``
    file), the simplices are guaranteed to be in chronological order.

    Parameters
    ----------
    folder : Path | str
        Path to the folder containing the dataset.

    Returns
    -------
    list[Simplex]
        List of simplices in the dataset.

    Raises
    ------
    ValueError
        If the folder does not exist or is not in the expected format.

    Notes
    -----
    - If the dataset has node labels, simplices have a ``label`` attribute.
    - If the dataset is temporal, simplices have a ``time`` attribute.
    """
    if not isinstance(folder, Path):
        folder = Path(folder)

    name = _validate_folder(folder)

    with (
        (folder / f"{name}-nverts.txt").open() as num_vertices_file,
        (folder / f"{name}-simplices.txt").open() as simplices_file,
    ):
        simplices = [
            Simplex([int(simplices_file.readline()) for _ in range(simplex_size)])
            for simplex_size in map(int, num_vertices_file)
        ]

    if (folder / f"{name}-times.txt").exists():
        with (folder / f"{name}-times.txt").open() as times_file:
            simplex_times = [int(line) for line in times_file]
        for time, simplex in zip(simplex_times, simplices, strict=True):
            simplex["time"] = time
        simplices = sorted(simplices, key=lambda simplex: simplex["time"])

    return simplices



"""Configure our testing suite."""

import networkx as nx
import numpy as np
import pytest

import toponetx as tnx


@pytest.fixture(autouse=True)
def doctest_default_imports(doctest_namespace):
    """Add default imports to the doctest namespace.

    This fixture adds the following default imports to every doctest, so that their use
    is consistent across all doctests without boilerplate imports polluting the
    doctests themselves:

    .. code-block:: python

        import numpy as np
        import networkx as nx
        import toponetx as tnx

    Parameters
    ----------
    doctest_namespace : dict
        The namespace of the doctest.
    """
    doctest_namespace["np"] = np
    doctest_namespace["nx"] = nx
    doctest_namespace["tnx"] = tnx



"""Read and write complexes as pickled objects."""

import pickle
from pathlib import Path

__all__ = ["load_from_pickle", "to_pickle"]


def to_pickle(obj, filename: str) -> None:
    """Write object to a pickle file.

    Parameters
    ----------
    obj : object
        Object to write.
    filename : Path or str
        Filename.
    """
    with Path(filename).open("wb") as f:
        pickle.dump(obj, f)


def load_from_pickle(filepath):
    """Return object from file.

    Parameters
    ----------
    filepath : Path or str
        Filepath.

    Returns
    -------
    object
        Object.
    """
    with Path(filepath).open("rb") as f:
        return pickle.load(f)  # noqa: S301



"""Test delaunay triangulation."""

import numpy as np

from toponetx.classes.simplex import Simplex
from toponetx.transform import delaunay_triangulation


def test_delaunay_triangulation_simple():
    """Test the Delaunay triangulation of a simple set of points."""
    points = np.array([[0, 0], [1, 0], [0, 1]])
    SC = delaunay_triangulation(points)

    assert set(SC.simplices) == {
        Simplex((0,)),
        Simplex((1,)),
        Simplex((2,)),
        Simplex((0, 1)),
        Simplex((0, 2)),
        Simplex((1, 2)),
        Simplex((0, 1, 2)),
    }


def test_delaunay_triangulation():
    """Test the Delaunay triangulation of a set of points."""
    points = np.array([[0, 0], [1, 0], [0, 1], [5, 5]])
    SC = delaunay_triangulation(points)

    assert set(SC.simplices) == {
        Simplex((0,)),
        Simplex((1,)),
        Simplex((2,)),
        Simplex((3,)),
        Simplex((0, 1)),
        Simplex((0, 2)),
        Simplex((1, 2)),
        Simplex((1, 3)),
        Simplex((2, 3)),
        Simplex((0, 1, 2)),
        Simplex((1, 2, 3)),
    }



"""Test the classic cell complex generators."""

from toponetx.generators.classic_cell_complexes import (
    pyrmaid_complex,
    single_cell_complex,
)


class TestSingleCellComplex:
    """Test the `single_cell_complex` function."""

    def test_single_cell(self):
        """Test `single_cell_complex` for zero nodes."""
        CC = single_cell_complex(6)
        assert len(CC.cells) == 1
        assert (1, 2, 3, 4, 5) in CC.cells


class TestPyrmaidComplex:
    """Test the `pyrmaid_complex` function."""

    def test_pyrmaid_complex(self):
        """Test `pyrmaid_complex` for zero nodes."""
        CC = pyrmaid_complex(6)
        assert len(CC.cells) == 6
        assert (1, 2, 3, 4, 5) in CC.cells



"""Test utils loading datasets."""

import toponetx.datasets.utils as data_utils


class TestUtils:
    """Test datasets utils."""

    def test_load_ppi(self):
        """Test that the ppi dataset loads correctly."""
        G = data_utils.load_ppi()
        assert len(G.nodes) == 20
        assert len(G.edges) == 101



"""Test the random simplicial complex generators."""

from toponetx.generators.random_simplicial_complexes import (
    linial_meshulam_complex,
    multiparameter_linial_meshulam_complex,
    random_clique_complex,
)


class TestLinialMeshulamComplex:
    """Test the `linial_meshulam_complex` function."""

    def test_zero_nodes(self):
        """Test `linial_meshulam_complex` for zero nodes."""
        SC = linial_meshulam_complex(0, 0.1)
        assert SC.shape == ()

    def test_zero_probability(self):
        """Test `linial_meshulam_complex` for zero probability."""
        SC = linial_meshulam_complex(10, 0.0)
        assert SC.shape == (10, 45)

    def test_one_probability(self):
        """Test `linial_meshulam_complex` for one probability."""
        SC = linial_meshulam_complex(10, 1.0)
        assert SC.shape == (10, 45, 120)


class TestRandomCliqueComplex:
    """Test the `random_clique_complex` function."""

    def test_zero_nodes(self):
        """Test `random_clique_complex` for zero nodes."""
        SC = random_clique_complex(0, 0.1)
        assert SC.shape == ()

    def test_zero_probability(self):
        """Test `random_clique_complex` for zero probability."""
        SC = random_clique_complex(10, 0.0)
        assert SC.shape == (10,)


class TestMultiparameterLinialMeshulamComplex:
    """Test the `multiparameter_linial_meshulam_complex` function."""

    def test_zero_nodes(self):
        """Test `multiparameter_linial_meshulam_complex` for zero nodes."""
        SC = multiparameter_linial_meshulam_complex(0, [1.0, 0.1])
        assert SC.shape == ()

    def test_one_probabilities(self):
        """Test `multiparameter_linial_meshulam_complex` for one probabilities."""
        SC = multiparameter_linial_meshulam_complex(10, [1.0])
        assert SC.shape == (10, 45)

        SC = multiparameter_linial_meshulam_complex(10, [1.0, 1.0])
        assert SC.shape == (10, 45, 120)

    def test_ignore_higher_cliques(self):
        """Test that `multiparameter_linial_meshulam_complex` ignores higher-order cliques.

        It is crucial that higher-order cliques are only added when all lower-order
        cliques have been added as well. Otherwise, the probabilities are not correct.
        For this, we artificially disallow triangles but force 4-cliques with
        probability one. Nonetheless, no 3-simplex should be present.
        """
        SC = multiparameter_linial_meshulam_complex(10, [1.0, 0.0, 1.0])
        assert SC.shape == (10, 45)



"""Test hyperedge class."""

import pytest

from toponetx.classes.hyperedge import HyperEdge


class TestHyperEdgeCases:
    """Test case for the HyperEdge class."""

    def test_hyperedge_creation(self):
        """Test creating a HyperEdge object."""
        elements = (1, 2, 3)
        hyperedge = HyperEdge(elements)
        assert len(hyperedge) == 3
        assert tuple(hyperedge) == elements

        elements = (1,)
        hyperedge = HyperEdge(elements)
        assert len(hyperedge) == 1
        assert tuple(hyperedge) == elements

        hyperedge = HyperEdge([1])
        assert len(hyperedge) == 1
        assert tuple(hyperedge) == (1,)

    def test_getitem_(self):
        """Test getitem method."""
        hyperedge = HyperEdge([1])
        hyperedge["weight"] = 1
        with pytest.raises(KeyError):
            _ = hyperedge["weightss"]

    def test_hyperedge_with_rank_and_attributes(self):
        """Test creating a HyperEdge object with rank and attributes."""
        elements = ("a", "b", "c")
        rank = 10
        attributes = {"color": "red", "weight": 2.5}
        hyperedge = HyperEdge(elements, rank=rank, **attributes)
        assert hyperedge.rank == rank
        assert hyperedge["color"] == attributes["color"]
        assert hyperedge["weight"] == attributes["weight"]

    def test_hyperedge_contains(self):
        """Test checking element containment in a HyperEdge."""
        elements = (1, 2, 3)
        hyperedge = HyperEdge(elements)
        assert 1 in hyperedge
        assert 4 not in hyperedge

    def test_hyperedge_iteration(self):
        """Test iterating over the elements of a HyperEdge."""
        elements = (1, 2, 3)
        hyperedge = HyperEdge(elements)
        assert tuple(hyperedge) == elements

    def test_hyperedge_representation(self):
        """Test the string representation of a HyperEdge."""
        elements = (1, 2, 3)
        hyperedge = HyperEdge(elements)
        expected_repr = (
            f"Nodes set: {tuple(hyperedge.elements)}, rank: {hyperedge.rank}"
        )
        assert isinstance(repr(hyperedge), str)
        assert repr(hyperedge) == expected_repr

    def test_hyperedge_str(self):
        """Test the string conversion of a HyperEdge."""
        elements = (1, 2, 3)
        attributes = {"color": "red", "weight": 2.5}
        hyperedge = HyperEdge(elements, **attributes)
        expected_str = (
            f"Nodes set: {tuple(hyperedge.elements)}, attrs: {hyperedge._attributes}"
        )
        assert isinstance(str(hyperedge), str)
        assert str(hyperedge) == expected_str

    def test_set_get_method(self):
        """Test set get methods."""
        hyperedge = HyperEdge((1, 2, 3))
        hyperedge["weight"] = 1
        assert hyperedge["weight"] == 1

    def test_rank(self):
        """Test rank."""
        he = HyperEdge([1, 4, 2], rank=4)
        assert he.rank == 4
        he = HyperEdge([1, 5, 2])
        assert he.rank is None

    def test_eq(self):
        """Test eq."""
        he1 = HyperEdge((1, 4, 2), rank=4)
        he2 = HyperEdge((1, 4, 2), rank=4)
        assert he1 == he2

        he1 = HyperEdge([1, 4, 2], rank=4)
        he2 = HyperEdge([1, 4, 2], rank=5)
        assert he1 != he2

        he1 = HyperEdge([1, 4, 2], rank=4)
        he2 = HyperEdge([1, 4, 2])
        assert he1 != he2

        he1 = HyperEdge([1, 4, 2], rank=4)
        he2 = HyperEdge([1, 4, 2], name="A")
        assert he1 != he2

        assert he1 != 1

    def test_hash(self):
        """Test HyperEdge hash."""
        he1 = HyperEdge((1, 4, 2), rank=4)
        he2 = HyperEdge((1, 4, 2), rank=4)
        assert hash(he1) == hash(he2)

        he1 = HyperEdge([1, 4, 2], rank=4)
        he2 = HyperEdge([1, 4, 2], rank=5)
        assert hash(he1) != hash(he2)

    def test_inite_method_non_hashabe(self):
        """Test non hashable."""
        with pytest.raises(TypeError) as exp_exception:
            HyperEdge([1, [1], 2], rank=4)
        assert (
            str(exp_exception.value) == "Every element of HyperEdge must be hashable."
        )

    def test_duplicates_elements(self):
        """Test duplicates_elements."""
        with pytest.raises(ValueError) as exp_exception:
            HyperEdge([1, 1, 2], rank=4)

        assert str(exp_exception.value) == "A hyperedge cannot contain duplicate nodes."



"""Test simplicial complex class."""

import re
from unittest.mock import Mock

import networkx as nx
import numpy as np
import pytest
from scipy.sparse import bmat

from toponetx.classes.combinatorial_complex import CombinatorialComplex
from toponetx.classes.simplex import Simplex
from toponetx.classes.simplicial_complex import SimplicialComplex
from toponetx.datasets.mesh import stanford_bunny

try:
    import hypernetx as hnx
except ImportError:
    hnx = None

try:
    import spharapy.trimesh as tm
except ImportError:
    tm = None


class TestSimplicialComplex:
    """Test SimplicialComplex class."""

    def test_iterable_simplices(self):
        """Test TypeError for simplices not iterable."""
        with pytest.raises(TypeError):
            SimplicialComplex(simplices=1)

    def test_shape_property(self):
        """Test shape property."""
        sc = SimplicialComplex([[1, 2, 3], [2, 3, 4], [0, 1]])
        assert sc.shape == (5, 6, 2)

        sc = SimplicialComplex()
        assert sc.shape == ()

        # make sure that empty dimensions are not part of the shape after removal
        sc = SimplicialComplex([[1, 2, 3], [2, 3, 4], [0, 1]])
        sc.remove_nodes([2, 3])
        assert sc.shape == (3, 1)

    def test_dim_property(self):
        """Test dim property."""
        sc = SimplicialComplex([[1, 2, 3], [2, 3, 4], [0, 1]])
        assert sc.dim == 2

        sc.remove_nodes([2, 3])
        assert sc.dim == 1

    def test_nodes_property(self):
        """Test nodes property."""
        sc = SimplicialComplex([[1, 2, 3], [2, 3, 4], [0, 1]])
        nodes = sc.nodes
        assert len(nodes) == 5
        assert [1] in nodes
        assert [8] not in nodes

    def test_simplices_property(self):
        """Test simplices property."""
        sc = SimplicialComplex([[1, 2, 3], [2, 3, 4], [0, 1]])
        assert len(sc.simplices) == 13
        assert [1, 2, 3] in sc.simplices
        assert [2, 3, 4] in sc.simplices
        assert [0, 1] in sc.simplices

    def test_is_maximal(self):
        """Test is_maximal method."""
        sc = SimplicialComplex([[1, 2, 3], [2, 3, 4], [0, 1]])
        assert sc.is_maximal([1, 2, 3])
        assert not sc.is_maximal([1, 2])
        assert not sc.is_maximal([3])

        with pytest.raises(ValueError):
            sc.is_maximal([1, 2, 3, 4])

    def test_get_maximal_simplices_of_simplex(self) -> None:
        """Test the get_maximal_simplices_of_simplex method."""
        SC = SimplicialComplex([(1, 2, 3), (2, 3, 4), (0, 1), (5,)])

        assert SC.get_maximal_simplices_of_simplex((1, 2, 3)) == {Simplex((1, 2, 3))}
        assert SC.get_maximal_simplices_of_simplex((1, 2)) == {Simplex((1, 2, 3))}
        assert SC.get_maximal_simplices_of_simplex((2, 3)) == {
            Simplex((1, 2, 3)),
            Simplex((2, 3, 4)),
        }
        assert SC.get_maximal_simplices_of_simplex((0,)) == {Simplex((0, 1))}
        assert SC.get_maximal_simplices_of_simplex((5,)) == {Simplex((5,))}

    def test_constructor_using_graph(self):
        """Test input a networkx graph in the constructor."""
        G = nx.Graph()
        G.add_edge(0, 1)
        G.add_edge(2, 5)
        G.add_edge(5, 4, weight=5)
        SC = SimplicialComplex(G)

        assert (0, 1) in SC.simplices
        assert (2, 5) in SC.simplices
        assert (5, 4) in SC.simplices
        assert SC.simplices[(5, 4)]["weight"] == 5

    def test_skeleton_raise_errors(self):
        """Test skeleton raises."""
        G = nx.Graph()
        G.add_edge(0, 1)
        G.add_edge(2, 5)
        G.add_edge(5, 4, weight=5)
        SC = SimplicialComplex(G)

        with pytest.raises(ValueError):
            SC.skeleton(-2)

        with pytest.raises(ValueError):
            SC.skeleton(2)

    def test_str(self):
        """Test str method."""
        G = nx.Graph()
        G.add_edge(0, 1)
        G.add_edge(2, 5)
        G.add_edge(5, 4, weight=5)
        SC = SimplicialComplex(G, name="graph complex")
        assert (
            str(SC)
        ) == f"Simplicial Complex with shape {SC.shape} and dimension {SC.dim}"

    def test_rep_str(self):
        """Test repr string."""
        G = nx.Graph()
        G.add_edge(0, 1)
        G.add_edge(2, 5)
        G.add_edge(5, 4, weight=5)
        SC = SimplicialComplex(G)
        assert (repr(SC)) == "SimplicialComplex()"

    def test_iter(self):
        """Test iter method."""
        SC = SimplicialComplex([[1, 2, 3], [2, 4], [5]])
        simplices = set(SC)
        assert simplices == {
            Simplex((1,)),
            Simplex((2,)),
            Simplex((3,)),
            Simplex((4,)),
            Simplex((5,)),
            Simplex((1, 2)),
            Simplex((1, 3)),
            Simplex((2, 3)),
            Simplex((2, 4)),
            Simplex((1, 2, 3)),
        }

    def test_getittem__(self):
        """Test __getitem__ methods."""
        SC = SimplicialComplex()
        SC.add_simplex((0, 1), weight=5)
        SC.add_simplex((1, 2, 3), heat=5)

        assert SC[(0, 1)]["weight"] == 5
        assert SC[(1, 2, 3)]["heat"] == 5
        with pytest.raises(KeyError):
            _ = SC[(1, 2, 3, 4, 5)]["heat"]

        SC[(0, 1)]["new"] = 10
        assert SC[(0, 1)]["new"] == 10

    def test_setting_simplex_attributes(self):
        """Test setting simplex attributes through a `SimplicialComplex` object."""
        G = nx.Graph()
        G.add_edge(0, 1)
        G.add_edge(2, 5)
        G.add_edge(5, 4, weight=5)
        SC = SimplicialComplex(G, name="graph complex")
        SC.add_simplex((1, 2, 3), heat=5)

        SC[(1, 2, 3)]["heat"] = 6
        assert SC[(1, 2, 3)]["heat"] == 6

        SC[(2, 5)]["heat"] = 1
        assert SC[(2, 5)]["heat"] == 1

        s = Simplex((1, 2, 3, 4), heat=1)
        SC.add_simplex(s)
        assert SC[(1, 2, 3, 4)]["heat"] == 1

        s = Simplex(("A",), heat=1)
        SC.add_simplex(s)
        assert SC["A"]["heat"] == 1

    def test_add_simplices_from(self):
        """Test add simplices from."""
        with pytest.raises(TypeError):
            SC = SimplicialComplex()
            SC.add_simplices_from(4)

    def test_add_node(self):
        """Test add node."""
        SC = SimplicialComplex()
        SC.add_node(9)
        assert 9 in SC
        with pytest.raises(TypeError):
            SC.add_node({1, 2})

        with pytest.raises(ValueError):
            s = Simplex((1, 2, 3, 4))
            SC.add_node(s)

        s = Simplex({1})  # singleton simplex
        SC.add_node(s)
        assert s in SC

        SC = SimplicialComplex()
        assert SC.dim == -1
        SC.add_node(9)
        assert SC.dim == 0

        SC = SimplicialComplex()
        assert SC.dim == -1
        SC.add_node(9)
        assert SC.dim == 0

    def test_add_simplex(self):
        """Test add_simplex method."""
        # create a SimplicialComplex object with no simplices
        SC = SimplicialComplex()

        # add a simplex using the add_simplex() method
        SC.add_simplex([1, 2, 3])

        # assert that the simplex was added correctly
        assert (1, 2, 3) in SC.simplices
        assert [1, 2, 3] in SC.simplices
        assert (1, 2) in SC.simplices
        assert (1, 3) in SC.simplices
        assert (2, 3) in SC.simplices
        assert (1,) in SC.simplices
        assert (2,) in SC.simplices
        assert (3,) in SC.simplices

        SC.add_simplex((4, 5, 6), heat=5)
        assert (4, 5, 6) in SC.simplices
        assert [4, 5, 6] in SC.simplices
        assert (4, 5) in SC.simplices
        assert (4, 6) in SC.simplices
        assert (5, 6) in SC.simplices
        assert (4,) in SC.simplices
        assert (5,) in SC.simplices
        assert (6,) in SC.simplices

        # check that provided attributes are stored correctly
        SC.add_simplex((1, 2), edge_flow=10)
        assert SC[(1, 2)]["edge_flow"] == 10
        SC.add_simplex(Simplex((1, 2), edge_flow=20))
        assert SC[(1, 2)]["edge_flow"] == 20
        SC.add_simplex(Simplex((2, 3), a=1, b=2), b=5)
        assert SC[(2, 3)]["a"] == 1
        assert SC[(2, 3)]["b"] == 5

        with pytest.raises(TypeError):
            SC.add_simplex(iter([1, 2]))

        # simplex cannot contain unhashable elements
        with pytest.raises(TypeError):
            SC.add_simplex([[1, 2], [2, 3]])

        # simplex cannot contain duplicate nodes
        with pytest.raises(ValueError):
            SC.add_simplex((1, 2, 2))

        # use of reserved attributes is not allowed
        with pytest.raises(ValueError):
            SC.add_simplex((1, 2, 3), is_maximal=True)
        with pytest.raises(ValueError):
            SC.add_simplex((1, 2, 3), membership={})

        # add hashable, non iterable node to SC
        SC.add_simplex(11)
        assert 11 in SC.simplices

        # add random string to SC
        SC.add_simplex("test")
        assert ("test",) in SC.simplices

    def test_contains(self):
        """Test the __contains__ method."""
        SC = SimplicialComplex([[1, 2], [3], [4]])

        assert 1 in SC
        assert 3 in SC
        assert 5 not in SC

        assert (1, 2) in SC
        assert (1, 3) not in SC

    def test_remove_maximal_simplex(self):
        """Test remove_maximal_simplex method."""
        SC = SimplicialComplex([[1, 2, 3], [2, 3, 4], [0, 1], [10]])
        SC.remove_maximal_simplex([1, 2, 3])
        SC.remove_maximal_simplex(10)

        assert (1, 2, 3) not in SC.simplices
        assert (10,) not in SC.simplices

        SC = SimplicialComplex([[1, 2, 3], [2, 3, 4]])
        SC.remove_maximal_simplex([2, 3, 4])

        assert [2, 3, 4] not in SC.simplices

        # check after the add_simplex method
        SC = SimplicialComplex()
        SC.add_simplex((1, 2, 3, 4), weight=1)
        c1 = Simplex((1, 2, 3, 4, 5))
        SC.add_simplex(c1)
        SC.remove_maximal_simplex((1, 2, 3, 4, 5))
        assert (1, 2, 3, 4, 5) not in SC.simplices

        # check removal with Simplex
        SC = SimplicialComplex()
        SC.add_simplex((1, 2, 3, 4), weight=1)
        c1 = Simplex((1, 2, 3, 4, 5))
        SC.add_simplex(c1)
        SC.remove_maximal_simplex(c1)
        assert (1, 2, 3, 4, 5) not in SC.simplices

        # check error when simplex not in complex
        with pytest.raises(ValueError):
            SC = SimplicialComplex()
            SC.add_simplex((1, 2, 3, 4), weight=1)
            SC.remove_maximal_simplex([5, 6, 7])

        # only maximal simplices can be removed
        with pytest.raises(ValueError):
            SC = SimplicialComplex()
            SC.add_simplex((1, 2, 3, 4), weight=1)
            SC.remove_maximal_simplex((1, 2, 3))

    def test_remove_nodes(self) -> None:
        """Test remove_nodes method."""
        SC = SimplicialComplex([[0, 1], [1, 2, 3], [2, 3, 4], [4, 5]])
        SC.remove_nodes([2, 5])

        assert [0, 1] in SC.simplices
        assert [1, 3] in SC.simplices
        assert [3, 4] in SC.simplices
        assert [4] in SC.simplices
        assert [2, 3] not in SC.simplices
        assert [2, 4] not in SC.simplices

        assert SC.is_maximal([0, 1])
        assert SC.is_maximal([1, 3])
        assert SC.is_maximal([3, 4])
        assert not SC.is_maximal([1])

    def test_skeleton_and_cliques(self):
        """Test skeleton and cliques methods."""
        G = nx.karate_club_graph()
        cliques = list(nx.enumerate_all_cliques(G))

        nodes = [i for i in cliques if len(i) == 1]
        edges = [i for i in cliques if len(i) == 2]
        faces = [i for i in cliques if len(i) == 3]
        threefaces = [i for i in cliques if len(i) == 4]
        fourfaces = [i for i in cliques if len(i) == 5]

        SC = SimplicialComplex(cliques)
        assert len(SC.skeleton(rank=0)) == len(nodes)
        assert len(SC.skeleton(rank=1)) == len(edges)
        assert len(SC.skeleton(rank=2)) == len(faces)
        assert len(SC.skeleton(rank=3)) == len(threefaces)
        assert len(SC.skeleton(rank=4)) == len(fourfaces)

    def test_incidence_matrix_1(self):
        """Test incidence_matrix shape and values."""
        SC = SimplicialComplex([[1, 2, 3], [2, 3, 4], [0, 1]])

        B2 = SC.incidence_matrix(rank=2)
        assert B2.shape == (6, 2)
        np.testing.assert_array_equal(
            B2.toarray(),
            np.array([[0, 1, -1, 1, 0, 0], [0, 0, 0, 1, -1, 1]]).T,
        )

        # repeat the same test, but with signed=False
        B2 = SC.incidence_matrix(rank=2, signed=False)
        assert B2.shape == (6, 2)
        np.testing.assert_array_equal(
            B2.toarray(),
            np.array([[0, 1, 1, 1, 0, 0], [0, 0, 0, 1, 1, 1]]).T,
        )

    def test_hodge_laplacian_matrix(self):
        """Test hodge_laplacian_matrix shape and values."""
        SC = SimplicialComplex([[1, 2, 3], [2, 3, 4], [0, 1]])
        L_hodge = SC.hodge_laplacian_matrix(rank=0)

        assert L_hodge.shape == (5, 5)

        D = np.diag([1, 3, 3, 3, 2])
        A = np.array(
            [
                [0.0, 1.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 0.0, 1.0, 1.0],
                [0.0, 1.0, 1.0, 0.0, 1.0],
                [0.0, 0.0, 1.0, 1.0, 0.0],
            ]
        )

        np.testing.assert_array_equal(L_hodge.toarray(), D - A)

    def test_hodge_laplacian_matrix_index(self):
        """Test unsigned hodge_laplacian_matrix method with index."""
        SC = SimplicialComplex([[1, 2, 3], [2, 3, 4], [0, 1]])
        row, L_hodge = SC.hodge_laplacian_matrix(rank=0, signed=False, index=True)

        assert L_hodge.shape == (5, 5)

        D = np.diag([1, 3, 3, 3, 2])
        A = np.array(
            [
                [0.0, 1.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 0.0, 1.0, 1.0],
                [0.0, 1.0, 1.0, 0.0, 1.0],
                [0.0, 0.0, 1.0, 1.0, 0.0],
            ]
        )

        np.testing.assert_array_equal(L_hodge.toarray(), np.abs(D - A))
        expected_row = {(0,): 0, (1,): 1, (2,): 2, (3,): 3, (4,): 4}

        assert row == expected_row

    def test_hodge_laplacian_matrix_rank_2(self):
        """Test unsigned hodge_laplacian_matrix method with index for different ranks."""
        SC = SimplicialComplex([[1, 2, 3], [2, 3, 4], [0, 1]])
        column, L_hodge = SC.hodge_laplacian_matrix(rank=2, signed=False, index=True)
        expected_col = {(1, 2, 3): 0, (2, 3, 4): 1}
        assert column == expected_col
        np.testing.assert_array_equal(L_hodge.toarray(), np.array([[3, 1], [1, 3]]))

    def test_hodge_laplacian_matrix_rank_1(self):
        """Test unsigned hodge_laplacian_matrix method with index for different ranks."""
        SC = SimplicialComplex([[1, 2, 3], [2, 3, 4], [0, 1]])
        column, L_hodge = SC.hodge_laplacian_matrix(rank=1, signed=False, index=True)
        expected_col = {
            (0, 1): 0,
            (1, 2): 1,
            (1, 3): 2,
            (2, 3): 3,
            (2, 4): 4,
            (3, 4): 5,
        }
        assert column == expected_col
        np.testing.assert_array_equal(
            L_hodge.toarray(),
            np.array(
                [
                    [2.0, 1.0, 1.0, 0.0, 0.0, 0.0],
                    [1.0, 3.0, 0.0, 0.0, 1.0, 0.0],
                    [1.0, 0.0, 3.0, 0.0, 0.0, 1.0],
                    [0.0, 0.0, 0.0, 4.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 3.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 3.0],
                ]
            ),
        )

        with pytest.raises(ValueError):
            SC.hodge_laplacian_matrix(rank=3)

    def test_adjacency_matrix(self):
        """Test adjacency_matrix shape and values."""
        SC = SimplicialComplex([[1, 2, 3], [2, 3, 4], [0, 1]])

        A = SC.adjacency_matrix(rank=0)

        assert A.shape == (5, 5)

        np.testing.assert_array_equal(
            A.toarray(),
            np.array(
                [
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 1.0, 1.0, 0.0],
                    [0.0, 1.0, 0.0, 1.0, 1.0],
                    [0.0, 1.0, 1.0, 0.0, 1.0],
                    [0.0, 0.0, 1.0, 1.0, 0.0],
                ]
            ),
        )

        ind, A = SC.adjacency_matrix(rank=0, index=True)
        expected_ind = {(0,): 0, (1,): 1, (2,): 2, (3,): 3, (4,): 4}
        assert ind == expected_ind

        with pytest.raises(ValueError):
            A = SC.adjacency_matrix(rank=0, weight=1)

    def test_get_boundaries(self):
        """Test the get_boundaries method."""
        simplices = [(1, 2, 3), (2, 3, 4), (0, 1)]
        with pytest.raises(TypeError):
            SimplicialComplex.get_boundaries(1)

        boundaries = SimplicialComplex.get_boundaries(simplices)
        assert frozenset((1, 2)) in boundaries
        assert frozenset((1, 3)) in boundaries
        assert frozenset((2, 3)) in boundaries

        # test for min dim/max dim combinations
        boundaries = SimplicialComplex.get_boundaries(simplices, min_dim=2)
        assert boundaries == {frozenset({1, 2, 3}), frozenset({2, 3, 4})}

        boundaries = SimplicialComplex.get_boundaries(simplices, min_dim=1, max_dim=1)
        assert boundaries == {
            frozenset({3, 4}),
            frozenset({2, 3}),
            frozenset({1, 2}),
            frozenset({0, 1}),
            frozenset({2, 4}),
            frozenset({1, 3}),
        }

        boundaries = SimplicialComplex.get_boundaries(simplices, max_dim=0)
        assert boundaries == {
            frozenset({2}),
            frozenset({3}),
            frozenset({1}),
            frozenset({4}),
            frozenset({0}),
        }

        # test for contradicting min dim/max dim combination
        boundaries = SimplicialComplex.get_boundaries(simplices, min_dim=2, max_dim=1)
        assert boundaries == set()

    def test_get_cofaces(self):
        """Test the get_cofaces method."""
        SC = SimplicialComplex()
        SC.add_simplex([1, 2, 3, 4])
        SC.add_simplex([1, 2, 4])
        SC.add_simplex([3, 4, 8])

        cofaces = SC.get_cofaces([1, 2, 4], codimension=1)
        cofaces = [simplex.elements for simplex in cofaces]
        assert (1, 2, 3, 4) in cofaces
        assert (3, 4, 8) not in cofaces
        # ... add more assertions based on the expected cofaces

    def test_get_star(self):
        """Test the get_star method."""
        SC = SimplicialComplex()
        SC.add_simplex([1, 2, 3, 4])
        SC.add_simplex([1, 2, 4])
        SC.add_simplex([3, 4, 8])

        star = SC.get_star([1, 2, 4])
        star = [simplex.elements for simplex in star]

        assert (1, 2, 4) in star
        assert (1, 2, 3, 4) in star
        # ... add more assertions based on the expected star

    def test_set_simplex_attributes(self):
        """Test the set_simplex_attributes method."""
        SC = SimplicialComplex()
        SC.add_simplex([1, 2, 3, 4])
        SC.add_simplex([1, 2, 4])
        SC.add_simplex([3, 4, 8])
        d = {(1, 2, 3): "red", (1, 2, 4): "blue"}
        SC.set_simplex_attributes(d, name="color")
        assert SC[(1, 2, 3)]["color"] == "red"

        # test for non-existing simplex
        d = {(3, 4, 5): "Nope"}
        SC.set_simplex_attributes(d, name="color")  # should not raise an error
        SC.set_simplex_attributes(d)  # should not raise an error

    def test_get_edges_from_matrix(self):
        """Test the get_edges_from_matrix method."""
        matrix = np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]])
        expected_edges = [(0, 1), (1, 0), (1, 2), (2, 1)]

        edges = SimplicialComplex().get_edges_from_matrix(matrix)

        assert set(edges) == set(expected_edges)

    def test_to_hasse_graph(self):
        """Test to hasse graph function."""
        SC = SimplicialComplex()
        SC.add_simplex([0, 1, 2])
        G = SC.to_hasse_graph()
        assert len(G.nodes) == 7
        assert len(G.edges) == 9

    def test_incidence_matrix(self):
        """Test incidence matrix."""
        SC = SimplicialComplex()
        SC.add_simplex([1, 2, 3, 4])
        SC.add_simplex([1, 2, 5])
        with pytest.raises(ValueError):
            SC.incidence_matrix(10)
        with pytest.raises(ValueError):
            SC.incidence_matrix(-1)

        B1 = SC.incidence_matrix(1)
        B2 = SC.incidence_matrix(2)
        assert B1.shape == tuple(SC.shape[:2])
        assert B2.shape == tuple(SC.shape[1:3])
        # B1 : 1->0
        # B2 : 2->1
        assert np.sum(abs(B1.dot(B2))) == 0  # boundary of boundary = 0
        # for the last to be true, order of bases must be canonical in both matrices

        row, col, B1 = SC.incidence_matrix(1, index=True)

        assert (len(row), len(col)) == B1.shape

        B0 = SC.incidence_matrix(0)
        assert B0.shape[-1] == len(SC.skeleton(0))

        assert np.sum(B0.dot(B1)) == 0  # boundary of boundary = 0

        # check incidence matrix for signed=False
        row, col, B1 = SC.incidence_matrix(1, index=True, signed=False)
        assert (len(row), len(col)) == B1.shape

        np.testing.assert_array_equal(
            B1.toarray(),
            np.array(
                [
                    [1, 1, 1, 1, 0, 0, 0, 0],
                    [1, 0, 0, 0, 1, 1, 1, 0],
                    [0, 1, 0, 0, 1, 0, 0, 1],
                    [0, 0, 1, 0, 0, 1, 0, 1],
                    [0, 0, 0, 1, 0, 0, 1, 0],
                ]
            ),
        )

    def test_coincidence_matrix_2(self):
        """Test coincidence matrix."""
        SC = SimplicialComplex()
        SC.add_simplex([1, 2, 3, 4])
        SC.add_simplex([1, 2, 4])
        B1T = SC.coincidence_matrix(1)
        B2T = SC.coincidence_matrix(2)
        assert B1T.shape == tuple(SC.shape[:2][::-1])
        assert B2T.shape == tuple(SC.shape[1:3][::-1])
        # B1T : 0->1
        # B2T : 1->2
        assert np.sum(abs(B2T.dot(B1T))) == 0  # coboundary of coboundary = 0
        # for the last to be true, order of bases must be canonical in both matrices

        row, col, B1T = SC.coincidence_matrix(1, index=True)

        assert (len(col), len(row)) == B1T.shape

        B0 = SC.coincidence_matrix(0)
        assert B0.shape[0] == len(SC.skeleton(0))

    def test_is_triangular_mesh(self):
        """Test is_triangular_mesh."""
        SC = stanford_bunny("simplicial")
        print("dim", SC.dim)
        print(list(SC.get_all_maximal_simplices()))
        assert SC.is_triangular_mesh()

        # test for non triangular mesh
        SC = SimplicialComplex()
        SC.add_simplex([1, 2, 3, 4])
        SC.add_simplex([1, 2, 4])
        SC.add_simplex([3, 4, 8])
        assert not SC.is_triangular_mesh()

    def test_to_trimesh(self):
        """Test to_trimesh."""
        SC = stanford_bunny("simplicial")
        trimesh_obj = SC.to_trimesh()
        assert len(trimesh_obj.vertices) == len(SC.skeleton(0))

        # test for non triangular mesh
        SC = SimplicialComplex()
        SC.add_simplex([1, 2, 3, 4])
        SC.add_simplex([1, 2, 4])
        SC.add_simplex([3, 4, 8])
        with pytest.raises(RuntimeError):
            SC.to_trimesh()

    @pytest.mark.skipif(
        tm is None, reason="Optional dependency 'spharapy' not installed."
    )
    def test_laplace_beltrami_operator(self):
        """Test laplace_beltrami_operator."""
        SC = stanford_bunny("simplicial")

        laplacian_matrix = SC.laplace_beltrami_operator()

        assert isinstance(laplacian_matrix, np.ndarray)

    @pytest.mark.skipif(
        tm is not None, reason="Optional dependency 'spharapy' installed."
    )
    def test_laplace_beltrami_operator_missing_dependency(self):
        """Test laplace_beltrami_operator for when `spharapy` is missing."""
        SC = stanford_bunny("simplicial")
        with pytest.raises(RuntimeError):
            SC.laplace_beltrami_operator()

    def test_from_nx_graph(self):
        """Test from networkx graph."""
        G = nx.Graph()
        G.add_edge(1, 2, weight=2)
        G.add_edge(3, 4, weight=4)
        SC = SimplicialComplex.from_nx(G)
        assert SC[(1, 2)]["weight"] == 2

    def test_is_connected(self):
        """Test is connected."""
        SC = stanford_bunny("simplicial")
        assert SC.is_connected()

    @pytest.mark.skipif(
        hnx is None, reason="Optional dependency 'hypernetx' not installed."
    )
    def test_simplicial_closure_of_hypergraph(self):
        """Test simplicial_closure_of_hypergraph."""
        hg = hnx.Hypergraph([[1, 2, 3, 4], [1, 2, 3]], static=True)
        sc = SimplicialComplex.simplicial_closure_of_hypergraph(hg)
        expected_simplices = [
            (1,),
            (2,),
            (3,),
            (4,),
            (1, 2),
            (1, 3),
            (1, 4),
            (2, 3),
            (2, 4),
            (3, 4),
            (1, 2, 3),
            (1, 2, 4),
            (1, 3, 4),
            (2, 3, 4),
            (1, 2, 3, 4),
        ]
        assert len(sc.simplices) == len(expected_simplices)

    @pytest.mark.skipif(
        hnx is None, reason="Optional dependency 'hypernetx' not installed."
    )
    def test_to_hypergraph(self):
        """Convert a SimplicialComplex to a hypergraph and compare the number of edges."""
        c1 = Simplex((1, 2, 3))
        c2 = Simplex((1, 2, 4))
        c3 = Simplex((2, 5))
        SC = SimplicialComplex([c1, c2, c3])
        expected_result = hnx.Hypergraph(
            {
                "e0": [1, 2],
                "e1": [1, 3],
                "e2": [1, 4],
                "e3": [2, 3],
                "e4": [2, 4],
                "e5": [2, 5],
                "e6": [1, 2, 3],
                "e7": [1, 2, 4],
            }
        )
        result = SC.to_hypergraph()
        assert len(result.edges) == len(expected_result.edges)

    @pytest.mark.skipif(
        hnx is not None, reason="Optional dependency 'hypernetx' installed."
    )
    def test_to_hypergraph_missing_dependency(self):
        """Test that `to_hypergraph` raises a correct error when `hypernetx` is not installed."""
        SC = SimplicialComplex()
        with pytest.raises(
            RuntimeError,
            match=re.escape(
                "Cannot transform simplicial complex to hypergraph, `hypernetx` is not installed."
            ),
        ):
            SC.to_hypergraph()

    def test_to_cell_complex(self):
        """Test to convert SimplicialComplex to Cell Complex."""
        SC = SimplicialComplex([(1, 2, 3), (1, 2, 4), (2, 5)])
        CC = SC.to_cell_complex()
        assert set(CC.nodes) == {1, 2, 3, 4, 5}
        assert set(CC.edges) == {(2, 1), (3, 1), (2, 3), (2, 4), (1, 4), (2, 5)}

    def test_to_combinatorial_complex(self):
        """Convert a SimplicialComplex to a CombinatorialComplex and compare the number of cells and nodes."""
        c1 = Simplex((1, 2, 3))
        c2 = Simplex((1, 2, 4))
        c3 = Simplex((2, 5))
        SC = SimplicialComplex([c1, c2, c3])
        expected_result = CombinatorialComplex()
        expected_result.add_cell((1, 2, 3), rank=2)
        expected_result.add_cell((1, 2, 4), rank=2)
        expected_result.add_cell((2, 5), rank=1)
        expected_result.add_cell((1, 2), rank=1)
        expected_result.add_cell((1, 3), rank=1)
        expected_result.add_cell((1, 4), rank=1)
        expected_result.add_cell((2, 3), rank=1)
        expected_result.add_cell((2, 4), rank=1)
        expected_result.add_cell((2, 5), rank=1)
        result = SC.to_combinatorial_complex()
        assert len(result.cells) == len(expected_result.cells)
        assert len(result.nodes) == len(expected_result.nodes)

    def test_from_gudhi(self):
        """Create a SimplicialComplex from a Gudhi SimplexTree and compare the number of simplices."""
        gudhi_simplices = [
            ([1, 2, 3, 5], 0.0),
            ([1, 2, 3], 0.0),
            ([1, 2, 5], 0.0),
            ([1, 2], 0.0),
            ([1, 3, 5], 0.0),
            ([1, 3], 0.0),
            ([1, 5], 0.0),
            ([1], 0.0),
            ([2, 3, 5], 0.0),
            ([2, 3], 0.0),
            ([2, 5], 0.0),
            ([2], 0.0),
            ([3, 5], 0.0),
            ([3], 0.0),
            ([5], 0.0),
        ]
        tree = Mock(["get_skeleton", "dimension"])
        tree.get_skeleton.side_effect = lambda i: (
            s for s in gudhi_simplices if len(s[0]) <= i + 1
        )
        tree.dimension.return_value = 3

        expected_result = SimplicialComplex()
        expected_result.add_simplex((1, 2, 3, 5))
        result = SimplicialComplex.from_gudhi(tree)
        assert len(result.simplices) == len(expected_result.simplices)

    def test_add_elements_from_nx_graph(self):
        """Add elements from a networkx graph to a SimplicialComplex and compare the number of simplices."""
        c1 = Simplex((1, 2, 3))
        c3 = Simplex((1, 2, 5))
        SC = SimplicialComplex([c1, c3])
        G = nx.Graph()
        G.add_edge(4, 5)
        expected_result = SimplicialComplex([c1, c3, Simplex((4, 5))])
        SC.add_elements_from_nx_graph(G)
        assert len(SC.simplices) == len(expected_result.simplices)

    def test_restrict_to_nodes(self):
        """Restrict a SimplicialComplex to the specified nodes and compare the result with the expected SimplicialComplex."""
        c1 = Simplex((1, 2, 3))
        c2 = Simplex((1, 2, 4))
        c3 = Simplex((1, 2, 5))
        SC = SimplicialComplex([c1, c2, c3])
        node_set = [1, 2, 3, 4]
        expected_result = SimplicialComplex(
            [
                Simplex((1,)),
                Simplex((2,)),
                Simplex((3,)),
                Simplex((4,)),
                Simplex((1, 2)),
                Simplex((1, 3)),
                Simplex((1, 4)),
                Simplex((2, 3)),
                Simplex((2, 4)),
                Simplex((1, 2, 3)),
                Simplex((1, 2, 4)),
            ]
        )
        result = SC.restrict_to_nodes(node_set)
        assert len(result.simplices) == len(expected_result.simplices)

    def test_get_all_maximal_simplices(self):
        """Retrieve all maximal simplices from a SimplicialComplex and compare the number of simplices."""
        SC = SimplicialComplex([(1, 2)])
        assert {simplex.elements for simplex in SC.get_all_maximal_simplices()} == {
            (1, 2)
        }

        SC = SimplicialComplex([(1, 2, 3), (1, 2, 4), (1, 2, 5)])
        result = list(SC.get_all_maximal_simplices())
        assert len(result) == 3

    def test_coincidence_matrix(self):
        """Test for coincidence matrix."""
        SC = SimplicialComplex()
        SC.add_simplex([0, 1, 2])

        _, _, B1 = SC.coincidence_matrix(1, index=True)

        assert B1.shape == (3, 3)
        assert np.allclose(
            B1.toarray(),
            np.array([[-1.0, 1.0, 0.0], [-1.0, 0.0, 1.0], [0.0, -1.0, 1.0]]),
        )

        B2 = SC.coincidence_matrix(2)

        assert B2.shape == (1, 3)
        assert np.allclose(B2.toarray(), np.array([[1.0, -1.0, 1.0]]))

    def test_down_laplacian_matrix(self):
        """Test the down_laplacian_matrix method of SimplicialComplex."""
        SC = SimplicialComplex()
        SC.add_simplex([1, 2, 3])
        SC.add_simplex([4, 5, 6])

        # Test case 1: Rank is within valid range
        result = SC.down_laplacian_matrix(rank=1, signed=True, weight=None, index=False)
        assert result.shape == (6, 6)

        # Test case 2: Rank is below the valid range
        with pytest.raises(ValueError):
            SC.down_laplacian_matrix(rank=0, signed=True, weight=None, index=False)

        # Test case 3: Rank is above the valid range
        with pytest.raises(ValueError):
            SC.down_laplacian_matrix(rank=5, signed=True, weight=None, index=False)

    def test_adjacency_matrix2(self):
        """Test the adjacency_matrix method of SimplicialComplex."""
        SC = SimplicialComplex()
        SC.add_simplex([1, 2, 3])
        SC.add_simplex([4, 5, 6])

        # Test case 1: Rank is within valid range
        result = SC.adjacency_matrix(rank=1, signed=True, weight=None, index=False)

        assert result.shape == (6, 6)

        # Test case 2: Rank is below the valid range
        with pytest.raises(ValueError):
            SC.adjacency_matrix(rank=-1, signed=False, weight=None, index=False)

        # Test case 3: Rank is above the valid range
        with pytest.raises(ValueError):
            SC.adjacency_matrix(rank=5, signed=False, weight=None, index=False)

    def test_dirac_operator_matrix(self):
        """Test dirac operator."""
        SC = SimplicialComplex()
        SC.add_simplex([1, 2, 3])
        SC.add_simplex([1, 2, 4])
        SC.add_simplex([3, 4, 8])
        m = SC.dirac_operator_matrix()
        size = sum(SC.shape)
        assert m.shape == (size, size)

        index, m = SC.dirac_operator_matrix(index=True)

        L = m.dot(m)

        check_L = bmat(
            [
                [SC.hodge_laplacian_matrix(0), None, None],
                [None, SC.hodge_laplacian_matrix(1), None],
                [None, None, SC.hodge_laplacian_matrix(2)],
            ]
        )

        assert np.linalg.norm((check_L - L).todense()) == 0

        assert (1,) in index
        assert (2,) in index
        assert (3,) in index
        assert (4,) in index
        assert len(index) == size

        index, m = SC.dirac_operator_matrix(index=True, signed=False)

        assert np.prod(m.todense() >= 0) == 1

        m = SC.dirac_operator_matrix(signed=False)

        assert np.prod(m.todense() >= 0) == 1

    def test_coadjacency_matrix(self):
        """Test the coadjacency_matrix method of SimplicialComplex."""
        SC = SimplicialComplex()
        SC.add_simplex([1, 2, 3])
        SC.add_simplex([4, 5, 6])
        # Test case 1: Rank is within valid range
        rank = 1
        signed = False
        weight = None
        index = False

        result = SC.coadjacency_matrix(rank, signed, weight, index)

        # Assert the result is of type scipy.sparse.csr.csr_matrix
        assert result.shape == (6, 6)

        # Test case 2: Rank is below the valid range
        rank = 0

        with pytest.raises(ValueError):
            SC.coadjacency_matrix(rank, signed, weight, index)

        # Test case 3: Rank is above the valid range
        rank = 5

        with pytest.raises(ValueError):
            SC.coadjacency_matrix(rank, signed, weight, index)

        # Test case 4: index is True
        # Test case 1: Rank is within valid range
        rank = 1
        signed = False
        weight = None
        index = True

        _, result = SC.coadjacency_matrix(rank, signed, weight, index)

        # Assert the result is of type scipy.sparse.csr.csr_matrix
        assert result.shape == (6, 6)

    def test_restrict_to_simplices(self):
        """Test the restrict_to_simplices method of SimplicialComplex."""
        c1 = Simplex((1, 2, 3))
        c2 = Simplex((1, 2, 4))
        c3 = Simplex((1, 2, 5))
        SC = SimplicialComplex([c1, c2, c3])
        SC1 = SC.restrict_to_simplices([c1, (2, 4)])
        assert len(SC1.simplices) == 9
        assert Simplex((1, 2, 3)) in SC1.simplices
        assert Simplex((2, 4)) in SC1.simplices
        assert Simplex((1, 2)) in SC1.simplices
        assert Simplex((1, 3)) in SC1.simplices
        assert Simplex((2, 3)) in SC1.simplices
        assert Simplex((1,)) in SC1.simplices
        assert Simplex((2,)) in SC1.simplices
        assert Simplex((3,)) in SC1.simplices
        assert Simplex((4,)) in SC1.simplices
        assert Simplex((1, 2, 3, 4)) not in SC1.simplices
        assert Simplex((1, 2, 5)) not in SC1.simplices

    def test_clone(self):
        """Test the clone method of SimplicialComplex."""
        SC = SimplicialComplex()
        SC.add_simplex([1, 2, 3], color="red")
        SC2 = SC.clone()
        assert SC2 is not SC
        assert SC2[(1, 2, 3)] is not SC[(1, 2, 3)]
        SC2.remove_maximal_simplex([1, 2, 3])
        assert 1 in SC
        assert (1, 2, 3) in SC.simplices

    def test_normalized_laplacian_matrix(self):
        """Test the normalized_laplacian_matrix method of SimplicialComplex."""
        SC = SimplicialComplex([[1, 2, 3], [2, 3, 5], [0, 1]])
        L = SC.normalized_laplacian_matrix(rank=1)
        assert np.allclose(
            L.toarray(),
            np.array(
                [
                    [0.5, -0.2236068, -0.2236068, 0.0, 0.0, 0.0],
                    [-0.2236068, 0.59999996, 0.0, 0.0, -0.2236068, 0.0],
                    [-0.2236068, 0.0, 0.59999996, 0.0, 0.0, -0.2236068],
                    [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, -0.2236068, 0.0, 0.0, 0.75, 0.0],
                    [0.0, 0.0, -0.2236068, 0.0, 0.0, 0.75],
                ]
            ),
        )

    @pytest.mark.skipif(
        tm is None, reason="Optional dependency 'spharapy' not installed."
    )
    def test_from_spharapy(self):
        """Test the from_spharapy method of SimplicialComplex (support for spharapy trimesh)."""
        mesh = tm.TriMesh(
            [[0, 1, 2]], [[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0]]
        )
        SC = SimplicialComplex.from_spharapy(mesh)
        simplices = SC.simplices
        assert len(simplices) == 7
        assert [0, 1, 2] in simplices
        assert [0, 1] in simplices
        assert [0, 2] in simplices
        assert [1, 2] in simplices
        assert [0] in simplices
        assert [1] in simplices
        assert [2] in simplices

    def test_graph_skeleton(self):
        """Test the graph_skeleton method of SimplicialComplex."""
        SC = SimplicialComplex(
            [
                (2, 6),
                (4, 5),
                (4, 7),
                (5, 6),
                (5, 7),
                (1, 2, 3),
                (2, 3, 4),
            ]
        )
        SC[1]["some_data"] = 1
        SC[(2, 6)]["some_data"] = 42

        G = SC.graph_skeleton()
        assert G.number_of_nodes() == 7
        assert G.number_of_edges() == 10
        assert G.nodes[1]["some_data"] == 1
        assert G.edges[(2, 6)]["some_data"] == 42



"""Test graph to cell complex transformation."""

import networkx as nx

from toponetx.transform.graph_to_cell_complex import homology_cycle_cell_complex


class TestGraphToCellComplex:
    """Test graph to cell complex transformation."""

    def test_homology_cycle_cell_complex(self):
        """Test homology_cycle_cell_complex."""
        cx = homology_cycle_cell_complex(nx.karate_club_graph())
        assert len(cx.cells) != 0



"""Test the data loading functions for Benson's datasets."""

from pathlib import Path

import pytest

from toponetx.classes.simplex import Simplex
from toponetx.datasets.benson import load_benson_hyperedges, load_benson_simplices


def test_load_benson_hyperedges_folder_error() -> None:
    """Test that the `load_benson_hyperedges` function raises an error for an invalid folder."""
    with pytest.raises(ValueError):
        load_benson_hyperedges("./")
    with pytest.raises(ValueError):
        load_benson_hyperedges("./nonexistent_folder")
    with pytest.raises(ValueError):
        load_benson_hyperedges("./test_benson.py")


def test_load_benson_hyperedges_dataset_2() -> None:
    """Test the `load_benson_hyperedges` function."""
    nodes, hyperedges = load_benson_hyperedges(
        Path(__file__).parent / "benson_sample_dataset-2"
    )
    assert len(nodes) == 7
    assert len(hyperedges) == 6

    for node in nodes:
        assert isinstance(node, Simplex)
        assert node["name"] == f"Node {next(iter(node))}"
        assert node["label"] in ["Label 1", "Label 2"]

    for simplex in hyperedges:
        assert isinstance(simplex, Simplex)


def test_load_benson_hyperedges_dataset_3() -> None:
    """Test the `load_benson_hyperedges` function."""
    nodes, hyperedges = load_benson_hyperedges(
        Path(__file__).parent / "benson_sample_dataset-3"
    )
    assert len(nodes) == 7
    assert len(hyperedges) == 6

    for node in nodes:
        assert isinstance(node, Simplex)
        assert node["name"] == f"Node {next(iter(node))}"
        assert isinstance(node["label"], list)
        assert all(
            label in ["Label 1", "Label 2", "Label 3"] for label in node["label"]
        )

    for simplex in hyperedges:
        assert isinstance(simplex, Simplex)


def test_load_benson_simplices_folder_error() -> None:
    """Test that the `load_benson_simplices` function raises an error for an invalid folder."""
    with pytest.raises(ValueError):
        load_benson_simplices("./")
    with pytest.raises(ValueError):
        load_benson_simplices("./nonexistent_folder")
    with pytest.raises(ValueError):
        load_benson_simplices("./test_benson.py")


def test_load_benson_simplices() -> None:
    """Test the `load_benson_simplices` function."""
    simplices = load_benson_simplices(Path(__file__).parent / "benson_sample_dataset")
    assert len(simplices) == 8

    for simplex in simplices:
        assert isinstance(simplex, Simplex)
        assert simplex["time"] in [1, 2]



"""Test distance measures module."""

import pytest

from toponetx.algorithms.distance_measures import (
    cell_diameter,
    cell_diameters,
    diameter,
    node_diameters,
)
from toponetx.classes.cell_complex import CellComplex
from toponetx.classes.colored_hypergraph import ColoredHyperGraph
from toponetx.classes.simplicial_complex import SimplicialComplex


class TestDistanceMeasures:
    """Test distance measures module."""

    def test_node_diameters(self):
        """Test for the node_diameters method."""
        CC = CellComplex()  # Initialize your class object

        # Add some cells to the complex
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([5, 6, 7], rank=2)

        # Test the function
        result = list(node_diameters(CC))
        expected_result = [[1, 1], [{2, 3, 4}, {5, 6, 7}]]
        assert result == expected_result

        CHG = ColoredHyperGraph([[1, 2, 3], [2, 3, 4]], ranks=2)
        CHG.add_cell([8, 9, 10], rank=2)
        result = list(node_diameters(CHG))
        expected_result = [
            [2, 1],
            [
                {frozenset({1}), frozenset({2}), frozenset({3}), frozenset({4})},
                {frozenset({8}), frozenset({9}), frozenset({10})},
            ],
        ]
        assert result == expected_result

    def test_cell_diameters(self):
        """Test for the cell_diameters method."""
        CC = CellComplex()  # Initialize your class object

        # Add some cells to the complex
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([5, 6, 7], rank=2)

        # Test the function
        result = list(cell_diameters(CC))
        expected_result = [
            [1, 1],
            [{(2, 3), (2, 3, 4), (2, 4), (3, 4)}, {(5, 6), (5, 6, 7), (5, 7), (6, 7)}],
        ]
        assert result == expected_result

        # invalid input
        with pytest.raises(TypeError):
            cell_diameters(SimplicialComplex())

    def test_diameter(self):
        """Test for the diameter method."""
        CC = CellComplex()  # Initialize your class object

        # Add some cells to the complex
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([5, 6, 7], rank=2)
        CC.add_cell([2, 5], rank=1)
        # Test the function
        result = diameter(CC)
        expected_result = 3
        assert result == expected_result

        # invalid input
        with pytest.raises(TypeError):
            diameter(SimplicialComplex())

        # diameter is not defined for non-connected complexes
        CC.add_cell([10, 22], rank=1)
        with pytest.raises(RuntimeError):
            _ = diameter(CC)

    def test_cell_diameter(self):
        """Test for the cell_diameter method."""
        CC = CellComplex()  # Initialize your class object

        # Add some cells to the complex
        CC.add_cell([2, 3, 4], rank=2)
        CC.add_cell([5, 6, 7], rank=2)
        CC.add_cell([2, 5], rank=1)

        # Test the function
        result = cell_diameter(CC)
        expected_result = 4
        assert result == expected_result

        # invalid input
        with pytest.raises(TypeError):
            cell_diameter(SimplicialComplex())

        # cell diameter is not defined for non-connected complexes
        CC.add_cell([10, 22], rank=1)
        with pytest.raises(RuntimeError):
            _ = cell_diameter(CC)



"""Test path complex class."""

import networkx as nx
import numpy as np
import pytest

from toponetx.classes.path import Path
from toponetx.classes.path_complex import PathComplex


class TestPathComplex:
    """Test path complex class."""

    def test_init_empty_path_complex(self):
        """Test empty path complex."""
        PC = PathComplex()
        assert len(PC.paths) == 0
        assert len(PC.nodes) == 0
        assert len(PC.edges) == 0
        assert PC.dim == -1

    def test_iterable_paths(self):
        """Test TypeError when paths is not iterable."""
        with pytest.raises(TypeError):
            _ = PathComplex(paths=1)

    def test_shape_property(self):
        """Test shape property."""
        PC = PathComplex([[1, 2, 3], [1, 2, 4]])
        assert PC.shape == (4, 3, 2)

        PC = PathComplex()
        assert PC.shape == ()

    def test_dim_property(self):
        """Test dim property."""
        PC = PathComplex([[1, 2, 3], [1, 2, 4]])
        assert PC.dim == 2

        PC = PathComplex()
        assert PC.dim == -1  # empty path complex

    def test_nodes_property(self):
        """Test nodes property."""
        PC = PathComplex([[1, 2, 3], [1, 2, 4]])
        nodes = PC.nodes
        assert len(nodes) == 4
        assert 1 in nodes
        assert 2 in nodes
        assert 6 not in nodes

    def test_edges_property(self):
        """Test edges property."""
        PC = PathComplex([[1, 2, 3], [1, 2, 4]])
        edges = PC.edges
        assert len(edges) == 3
        assert (1, 2) in edges
        assert (2, 3) in edges
        assert (2, 4) in edges
        assert (1, 3) not in edges

    def test_path_property(self):
        """Test path property."""
        PC = PathComplex([[1, 2, 3], [1, 2, 4]])
        assert [1, 2, 3] in PC.paths
        assert [1, 2, 4] in PC.paths
        assert [1, 2] in PC.paths
        assert 4 in PC.paths

    def test_add_path(self):
        """Test add path."""
        PC = PathComplex()
        PC.add_path([1, 2, 3])
        assert [1, 2, 3] in PC.paths
        assert [1, 2] in PC.paths
        assert 3 in PC.paths

        with pytest.raises(ValueError):
            PC.add_path([3, 2])

        with pytest.raises(ValueError):
            PC.add_path([3, 2, 2])

        with pytest.raises(ValueError):
            PC.add_path([3, 2, 3])

        PC = PathComplex(reserve_sequence_order=True)
        PC.add_path([3, 2, 1])
        assert [3, 2, 1] in PC.paths

        PC = PathComplex()
        PC.add_path([1, 2])
        PC.add_path([2, 3])
        assert (
            [1, 2, 3] not in PC.paths
        )  # because manually adding subpaths does not add superpath. Future features?
        assert [1, 2] in PC.paths
        assert [2, 3] in PC.paths

    def test_contains(self):
        """Test the __contains__ method."""
        PC = PathComplex([[1, 2], [3], [4]])

        assert 1 in PC
        assert 3 in PC
        assert 5 not in PC

        assert (1, 2) in PC
        assert (1, 3) not in PC

    def test_constructor_using_graph(self):
        """Test constructor using graph."""
        G = nx.Graph()
        G.add_edge(1, 2)
        G.add_edge(2, 3)
        PC = PathComplex(G)

        assert [1, 2, 3] in PC.paths  # because constructing from graph
        assert [1, 2] in PC.paths
        assert [2, 3] in PC.paths
        assert 1 in PC.paths
        assert 2 in PC.paths
        assert 3 in PC.paths

        PC = PathComplex(G, allowed_paths=[[1], [2], [3], [1, 2]])
        assert [2, 3] in PC.paths
        assert PC._allowed_paths == {(1,), (2,), (3,), (1, 2), (2, 3), (1, 2, 3)}

        PC = PathComplex(G, allowed_paths=[])
        assert [
            1,
            2,
            3,
        ] in PC.paths  # because constructing from graph and allowed_paths is empty
        assert [1, 2] in PC.paths
        assert [2, 3] in PC.paths
        assert 1 in PC.paths
        assert 2 in PC.paths
        assert 3 in PC.paths

        G = nx.Graph()
        G.add_edge(3, 2)
        PC = PathComplex(G)
        assert [2, 3] in PC.paths
        assert [3, 2] not in PC.paths

    def test_clone(self):
        """Test clone."""
        PC = PathComplex([[1, 2, 3], [1, 2, 4]])
        PX_clone = PC.clone()
        assert set(PX_clone.paths) == set(PC.paths)
        assert PX_clone is not PC

    def test_skeleton(self):
        """Test skeleton."""
        PC = PathComplex(
            [
                [0, 1],
                [1, 2, 3],
                [1, 3, 2],
                [2, 1, 3],
                [0, 1, 2],
                [0, 1, 3],
                [0, 1, 2, 3],
                [0, 1, 3, 2],
            ],
            max_rank=2,
        )
        assert PC.skeleton(0) == [(0,), (1,), (2,), (3,)]
        assert PC.skeleton(1) == [(0, 1), (1, 2), (1, 3), (2, 3)]
        assert PC.skeleton(2) == [(0, 1, 2), (0, 1, 3), (1, 2, 3), (1, 3, 2), (2, 1, 3)]
        with pytest.raises(ValueError):
            PC.skeleton(3)

    def test_skeleton_raise_errors(self):
        """Test skeleton raise errors."""
        PC = PathComplex([[1, 2, 3], [1, 3]])
        with pytest.raises(ValueError):
            PC.skeleton(-1)

        with pytest.raises(TypeError):
            PC.skeleton(1.5)

        with pytest.raises(TypeError):
            PC.skeleton(1, 2)

        with pytest.raises(ValueError):
            PC.skeleton(3)

    def test_str(self):
        """Test str."""
        PC = PathComplex([[1, 2, 3], [1, 2, 4]], name="PC")
        assert str(PC) == "Path Complex with shape (4, 3, 2) and dimension 2"

    def test_add_str_nodes(self):
        """Test add str nodes."""
        PC = PathComplex([[1, 2, 3], [2, 4, "a"]])
        assert "a" in PC.paths
        assert "a" in PC.nodes
        assert [4, "a"] in PC.edges

        with pytest.raises(ValueError):
            PC.add_path(["a", 1, 2])

    def test_repr_str(self):
        """Test repr str."""
        PC = PathComplex([[1, 2, 3], [1, 2, 4]])
        assert (repr(PC)) == "PathComplex()"

    def test_getittem_set_attributes(self):
        """Test getitem and set_attributes methods."""
        PC = PathComplex([[1, 2, 3], [1, 2, 4]])
        assert PC[[1, 2, 3]] == {}
        assert PC[1] == {}

        with pytest.raises(KeyError):
            PC[5]

        PC.add_path([1, 2, 5], heat=55)
        assert PC[[1, 2, 5]] == {"heat": 55}

        PC.set_path_attributes({(1, 2, 5): 66}, "heat")
        assert PC[[1, 2, 5]] == {"heat": 66}

        PC.set_path_attributes({(1, 2, 5): {"heat": 77, "color": "red"}})
        assert PC[[1, 2, 5]] == {"heat": 77, "color": "red"}

        PC.set_path_attributes({1: 88, (1, 2, 5): 99}, "heat")
        assert PC[1] == {"heat": 88}
        assert PC.nodes[1] == {"heat": 88}
        assert PC[[1, 2, 5]] == {"heat": 99, "color": "red"}

        PC.set_path_attributes({(1, 2): {"heat": 100, "color": "red"}})
        assert PC[[1, 2]] == {"heat": 100, "color": "red"}
        assert PC.edges[[1, 2]] == {"heat": 100, "color": "red"}

        PC.set_path_attributes({(1,): {"heat": 200, "color": "red"}})
        assert PC[[1]] == {"heat": 200, "color": "red"}
        assert PC.nodes[1] == {"heat": 200, "color": "red"}
        assert PC[1] == {"heat": 200, "color": "red"}

        PC.set_path_attributes({1: 90, 4: 23}, "heat")
        assert PC[1] == {"heat": 90, "color": "red"}
        assert PC.nodes[1] == {"heat": 90, "color": "red"}
        assert PC[4] == {"heat": 23}

        PC.set_path_attributes({(1, 2): 882, (1, 2, 5): 939}, "heat")
        assert PC[[1, 2]] == {"heat": 882, "color": "red"}
        assert PC.edges[[1, 2]] == {"heat": 882, "color": "red"}
        assert PC[[1, 2, 5]] == {"heat": 939, "color": "red"}

        PC.add_path(6, heat=77)
        assert PC[6] == {"heat": 77}
        assert PC.nodes[6] == {"heat": 77}

        PC.set_node_attributes({6: 88, 5: 22}, "heat")
        assert PC[6] == {"heat": 88}
        assert PC.nodes[6] == {"heat": 88}
        assert PC[5] == {"heat": 22}
        assert PC.nodes[5] == {"heat": 22}

        PC.set_node_attributes({6: 99, (5,): 33}, "heat")
        assert PC[6] == {"heat": 99}
        assert PC.nodes[6] == {"heat": 99}
        assert PC[5] == {"heat": 33}
        assert PC.nodes[5] == {"heat": 33}

        PC.set_node_attributes({6: {"heat": 100, "color": "red"}, (5,): {"heat": 34}})
        assert PC[6] == {"heat": 100, "color": "red"}
        assert PC.nodes[6] == {"heat": 100, "color": "red"}
        assert PC[5] == {"heat": 34}
        assert PC.nodes[5] == {"heat": 34}

        PC.add_path([4, 5], heat=77)
        assert PC[[4, 5]] == {"heat": 77}
        assert PC.edges[(4, 5)] == {"heat": 77}

        PC.set_edge_attributes({(4, 5): 88}, "heat")
        assert PC[[4, 5]] == {"heat": 88}
        assert PC.edges[(4, 5)] == {"heat": 88}

        PC.set_edge_attributes({(4, 5): {"heat": 99, "color": "red"}})
        assert PC[[4, 5]] == {"heat": 99, "color": "red"}
        assert PC.edges[(4, 5)] == {"heat": 99, "color": "red"}

        with pytest.raises(ValueError):
            PC.set_node_attributes({(4, 5): 55}, name="heat")
        with pytest.raises(ValueError):
            PC.set_node_attributes({(4, 5): {"heat": 55}})
        with pytest.raises(TypeError):
            PC.set_node_attributes({(4,): [5]})
        with pytest.raises(ValueError):
            PC.set_edge_attributes({(4, 5, 6): 55}, name="heat")
        with pytest.raises(ValueError):
            PC.set_edge_attributes({(4, 5, 6): {"heat": 55}})
        with pytest.raises(TypeError):
            PC.set_edge_attributes({(4, 5): [5]})
        with pytest.raises(TypeError):
            PC.set_path_attributes({(4, 5): [5]})

    def test_get_len_(self):
        """Test get size of the path complex."""
        PC = PathComplex()
        PC.add_paths_from([[1, 2, 3], [1, 2, 4]])
        assert len(PC) == 9

    def test_add_paths_from(self):
        """Test add paths from graph."""
        PC = PathComplex()
        PC.add_paths_from([[1, 2, 3], [1, 2, 4]])
        assert [1, 2, 3] in PC.paths
        assert [1, 2] in PC.paths
        assert [2, 3] in PC.paths
        assert 1 in PC.paths
        assert 2 in PC.paths
        assert 3 in PC.paths
        assert 1 in PC.nodes
        assert 2 in PC.nodes
        assert 3 in PC.nodes
        assert 4 in PC.nodes
        assert [1, 2] in PC.edges
        assert [2, 4] in PC.edges

    def test_add_node(self):
        """Test add node."""
        PC = PathComplex()
        PC.add_node(1)
        assert 1 in PC.paths
        assert [1] in PC.paths
        assert 1 in PC.nodes

        with pytest.raises(TypeError):
            PC.add_node([1, 2])

        PC.add_node(Path(2))
        assert 2 in PC.paths
        assert [2] in PC.paths
        assert 2 in PC.nodes

        with pytest.raises(ValueError):
            PC.add_node(Path([2, 3]))

    def test_remove_nodes(self):
        """Test remove_nodes method."""
        PC = PathComplex([[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3]])
        assert 1 in PC.nodes
        PC.remove_nodes([1])
        assert 1 not in PC.paths
        assert [1, 2] not in PC.paths
        assert [1, 3] not in PC.paths
        assert 1 not in PC.nodes
        assert [1, 2] not in PC.edges
        assert [1, 3] not in PC.edges

    def test_incidence_matrix(self):
        """Test incidence matrix."""
        PC = PathComplex(
            [[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3], [0, 1, 2], [0, 1, 3]]
        )
        assert np.all(PC.incidence_matrix(0).todense == np.zeros((0, len(PC.nodes))))
        assert np.all(
            PC.incidence_matrix(1, signed=False).todense()
            == np.array([[1, 0, 0, 0], [1, 1, 1, 0], [0, 1, 0, 1], [0, 0, 1, 1]])
        )
        assert np.all(
            PC.incidence_matrix(2, signed=False).todense()
            == np.array(
                [[1, 1, 0, 0, 0], [1, 0, 1, 1, 1], [0, 1, 1, 1, 1], [0, 0, 1, 1, 1]]
            )
        )

        with pytest.raises(ValueError):
            PC.incidence_matrix(-1)

        with pytest.raises(ValueError):
            PC.incidence_matrix(3)

        row, col, B = PC.incidence_matrix(0, index=True)
        assert row == {}
        assert col == {(0,): 0, (1,): 1, (2,): 2, (3,): 3}
        assert np.all(B.todense() == np.zeros((0, len(PC.nodes))))
        row, col, B = PC.incidence_matrix(0, index=True, signed=False)
        assert row == {}
        assert col == {(0,): 0, (1,): 1, (2,): 2, (3,): 3}
        assert np.all(B.todense() == np.zeros((0, len(PC.nodes))))

        row, col, B = PC.incidence_matrix(1, index=True)
        assert row == {(0,): 0, (1,): 1, (2,): 2, (3,): 3}
        assert col == {(0, 1): 0, (1, 2): 1, (1, 3): 2, (2, 3): 3}
        assert np.all(
            B.todense()
            == np.array([[-1, 0, 0, 0], [1, -1, -1, 0], [0, 1, 0, -1], [0, 0, 1, 1]])
        )
        row, col, B = PC.incidence_matrix(1, index=True, signed=False)
        assert row == {(0,): 0, (1,): 1, (2,): 2, (3,): 3}
        assert col == {(0, 1): 0, (1, 2): 1, (1, 3): 2, (2, 3): 3}
        assert np.all(
            B.todense()
            == np.array([[1, 0, 0, 0], [1, 1, 1, 0], [0, 1, 0, 1], [0, 0, 1, 1]])
        )

        row, col, B = PC.incidence_matrix(2, index=True)
        assert row == {(0, 1): 0, (1, 2): 1, (1, 3): 2, (2, 3): 3}
        assert col == {
            (0, 1, 2): 0,
            (0, 1, 3): 1,
            (1, 2, 3): 2,
            (1, 3, 2): 3,
            (2, 1, 3): 4,
        }
        assert np.all(
            B.todense()
            == np.array(
                [[1, 1, 0, 0, 0], [1, 0, 1, -1, 1], [0, 1, -1, 1, 1], [0, 0, 1, 1, -1]]
            )
        )
        row, col, B = PC.incidence_matrix(2, index=True, signed=False)
        assert row == {(0, 1): 0, (1, 2): 1, (1, 3): 2, (2, 3): 3}
        assert col == {
            (0, 1, 2): 0,
            (0, 1, 3): 1,
            (1, 2, 3): 2,
            (1, 3, 2): 3,
            (2, 1, 3): 4,
        }
        assert np.all(
            B.todense()
            == np.array(
                [[1, 1, 0, 0, 0], [1, 0, 1, 1, 1], [0, 1, 1, 1, 1], [0, 0, 1, 1, 1]]
            )
        )

    def test_up_laplacian_matrix(self):
        """Test up laplacian matrix."""
        PC = PathComplex(
            [[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3], [0, 1, 2], [0, 1, 3]]
        )
        with pytest.raises(ValueError):
            PC.up_laplacian_matrix(-1)
        with pytest.raises(ValueError):
            PC.up_laplacian_matrix(2)
        with pytest.raises(ValueError):
            PC.up_laplacian_matrix(1, weight="weight")

        row, L_up = PC.up_laplacian_matrix(0, index=True)
        row_B, _, B = PC.incidence_matrix(1, index=True)
        L_up_tmp = B.todense() @ B.todense().T

        assert np.all(L_up.todense() == L_up_tmp)
        assert row == row_B

        L_up = PC.up_laplacian_matrix(1)
        B = PC.incidence_matrix(2)
        L_up_tmp = B.todense() @ B.todense().T

        assert np.all(L_up.todense() == L_up_tmp)

        L_up = PC.up_laplacian_matrix(1, signed=True)
        B = PC.incidence_matrix(2, signed=True)
        L_up_tmp = B.todense() @ B.todense().T

        assert np.all(L_up.todense() == L_up_tmp)

    def test_down_laplacian_matrix(self):
        """Test down laplacian matrix."""
        PC = PathComplex(
            [[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3], [0, 1, 2], [0, 1, 3]]
        )
        with pytest.raises(ValueError):
            PC.down_laplacian_matrix(0)
        with pytest.raises(ValueError):
            PC.down_laplacian_matrix(3)
        with pytest.raises(ValueError):
            PC.down_laplacian_matrix(1, weight="weight")

        row, L_down = PC.down_laplacian_matrix(1, index=True)
        row_B, _, B = PC.incidence_matrix(1, index=True)
        L_down_tmp = B.todense().T @ B.todense()

        assert np.all(L_down.todense() == L_down_tmp)
        assert row == row_B

        L_down = PC.down_laplacian_matrix(2)
        B = PC.incidence_matrix(2)
        L_down_tmp = B.todense().T @ B.todense()

        assert np.all(L_down.todense() == L_down_tmp)

        L_down = PC.down_laplacian_matrix(2, signed=True)
        B = PC.incidence_matrix(2, signed=True)
        L_down_tmp = B.todense().T @ B.todense()

        assert np.all(L_down.todense() == L_down_tmp)

    def test_hodge_laplacian_matrix(self):
        """Test hodge laplacian matrix."""
        PC = PathComplex(
            [[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3], [0, 1, 2], [0, 1, 3]]
        )

        assert np.all(
            PC.hodge_laplacian_matrix(0).todense()
            == PC.up_laplacian_matrix(0).todense()
        )
        assert np.all(
            PC.hodge_laplacian_matrix(2).todense()
            == PC.down_laplacian_matrix(2).todense()
        )
        assert np.all(
            PC.hodge_laplacian_matrix(1).todense()
            == (PC.up_laplacian_matrix(1) + PC.down_laplacian_matrix(1)).todense()
        )

        assert np.all(
            PC.hodge_laplacian_matrix(0, signed=False).todense()
            == PC.up_laplacian_matrix(0, signed=False).todense()
        )
        assert np.all(
            PC.hodge_laplacian_matrix(2, signed=False).todense()
            == PC.down_laplacian_matrix(2, signed=False).todense()
        )
        assert np.all(
            PC.hodge_laplacian_matrix(1, signed=False).todense()
            == abs(
                PC.up_laplacian_matrix(1, signed=True)
                + PC.down_laplacian_matrix(1, signed=True)
            ).todense()
        )

        row, _ = PC.hodge_laplacian_matrix(0, index=True)
        assert row == PC.incidence_matrix(0, index=True)[1]

        row, _ = PC.hodge_laplacian_matrix(1, index=True)
        assert row == PC.incidence_matrix(1, index=True)[1]

        row, _ = PC.hodge_laplacian_matrix(2, index=True)
        assert row == PC.incidence_matrix(2, index=True)[1]

        with pytest.raises(ValueError):
            PC.hodge_laplacian_matrix(-1)

        with pytest.raises(ValueError):
            PC.hodge_laplacian_matrix(3)

    def test_adjacency_matrix(self):
        """Test adjacency matrix."""
        PC = PathComplex(
            [[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3], [0, 1, 2], [0, 1, 3]]
        )
        with pytest.raises(ValueError):
            PC.adjacency_matrix(-1)
        with pytest.raises(ValueError):
            PC.adjacency_matrix(2)

        adj_0 = PC.adjacency_matrix(0, signed=False).todense()
        assert np.all(
            adj_0 == np.array([[0, 1, 0, 0], [1, 0, 1, 1], [0, 1, 0, 1], [0, 1, 1, 0]])
        )

        adj_1 = PC.adjacency_matrix(1, signed=False).todense()

        assert np.all(
            adj_1 == np.array([[0, 1, 1, 0], [1, 0, 1, 1], [1, 1, 0, 1], [0, 1, 1, 0]])
        )

        row, _ = PC.adjacency_matrix(0, index=True)
        assert row == PC.incidence_matrix(0, index=True)[1]

        row, _ = PC.adjacency_matrix(1, index=True)
        assert row == PC.incidence_matrix(1, index=True)[1]

    def test_coadjacency_matrix(self):
        """Test coadjacency matrix."""
        PC = PathComplex(
            [[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3], [0, 1, 2], [0, 1, 3]]
        )
        with pytest.raises(ValueError):
            PC.coadjacency_matrix(0)
        with pytest.raises(ValueError):
            PC.coadjacency_matrix(3)

        coadj_1 = PC.coadjacency_matrix(1, signed=False).todense()
        assert np.all(
            coadj_1
            == np.array([[0, 1, 1, 0], [1, 0, 1, 1], [1, 1, 0, 1], [0, 1, 1, 0]])
        )

        coadj_2 = PC.coadjacency_matrix(2, signed=False).todense()
        assert np.all(
            coadj_2
            == np.array(
                [
                    [0, 1, 1, 1, 1],
                    [1, 0, 1, 1, 1],
                    [1, 1, 0, 1, 1],
                    [1, 1, 1, 0, 1],
                    [1, 1, 1, 1, 0],
                ]
            )
        )

        row, _ = PC.coadjacency_matrix(1, index=True)
        assert row == PC.incidence_matrix(0, index=True)[1]
        assert row == PC.incidence_matrix(1, index=True)[0]

        row, _ = PC.coadjacency_matrix(2, index=True)
        assert row == PC.incidence_matrix(1, index=True)[1]
        assert row == PC.incidence_matrix(2, index=True)[0]

    def test_restrict_to_nodes(self):
        """Test restrict_to_nodes."""
        PC = PathComplex(
            [[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3], [0, 1, 2], [0, 1, 3]], max_rank=3
        )
        res = PC.restrict_to_nodes([0, 1, 2])
        assert len(res.paths) == 6
        assert len(res.nodes) == 3
        assert len(res.edges) == 2

        res = PC.restrict_to_nodes([1, 2, 3])
        assert len(res.paths) == 9
        assert len(res.nodes) == 3
        assert len(res.edges) == 3

        with pytest.raises(ValueError):
            PC.restrict_to_nodes([])

    def test_restrict_to_paths(self):
        """Test restrict_to_paths."""
        PC = PathComplex(
            [[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3], [0, 1, 2], [0, 1, 3]], max_rank=3
        )

        res = PC.restrict_to_paths([[0, 1], [1, 2, 3]])
        assert len(res.paths) == 8
        assert len(res.nodes) == 4
        assert len(res.edges) == 3

        with pytest.raises(ValueError):
            PC.restrict_to_paths([])

    def test_get_node_attributes(self):
        """Test get_node_attributes."""
        PC = PathComplex()
        PC.add_node(0)
        PC.add_node(1, heat=55)
        PC.add_node(2, heat=66)
        PC.add_node(3, color="red")
        PC.add_node(2, color="blue")
        PC.add_paths_from([[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3]])

        assert PC.get_node_attributes("heat") == {(1,): 55, (2,): 66}
        assert PC.get_node_attributes("color") == {(2,): "blue", (3,): "red"}

    def test_get_edge_attributes(self):
        """Test get_edge_attributes."""
        PC = PathComplex()
        PC.add_paths_from([[0, 1], [1, 2, 3], [1, 3, 2], [2, 1, 3]])
        PC.add_path([0, 1], weight=32)
        PC.add_path([1, 2], weight=98)
        PC.add_path([1, 3], color="red")
        PC.add_path([2, 3], color="blue")

        assert PC.get_edge_attributes("weight") == {(0, 1): 32, (1, 2): 98}
        assert PC.get_edge_attributes("color") == {(1, 3): "red", (2, 3): "blue"}

    def test_get_path_attributes(self):
        """Test get_path_attributes."""
        PC = PathComplex()
        PC.add_paths_from([[0, 1]])
        PC.add_path([0, 1, 2], weight=43)
        PC.add_path([0, 1, 3], weight=98)
        PC.add_path([1, 2, 3], color="red")
        PC.add_path([1, 3, 2], color="blue")
        PC.add_path([2, 1, 3], color="green")

        assert PC.get_path_attributes("weight") == {(0, 1, 2): 43, (0, 1, 3): 98}
        assert PC.get_path_attributes("color") == {
            (1, 2, 3): "red",
            (1, 3, 2): "blue",
            (2, 1, 3): "green",
        }



"""Methods to compute the Delaunay triangulation of a set of points."""

import numpy as np
from scipy.spatial import Delaunay

from toponetx.classes import SimplicialComplex


def delaunay_triangulation(points: np.ndarray) -> SimplicialComplex:
    """
    Compute the Delaunay triangulation of a set of points in the plane.

    The resulting simplicial complex has nodes 0 to (n-1), where n is the number of
    points following the coordinate order in the input array.

    Parameters
    ----------
    points : np.ndarray
        An array of shape (n, 2) containing the coordinates of the points.

    Returns
    -------
    SimplicialComplex
        The Delaunay triangulation as a SimplicialComplex object.
    """
    triangles = Delaunay(points)
    return SimplicialComplex(triangles.simplices)



"""Module to distance measures on topological domains."""

from collections.abc import Hashable

import networkx as nx

from toponetx.classes.cell_complex import CellComplex
from toponetx.classes.colored_hypergraph import ColoredHyperGraph
from toponetx.classes.combinatorial_complex import CombinatorialComplex

__all__ = ["cell_diameter", "cell_diameters", "diameter", "node_diameters"]

# In this module, only cell complexes, combinatorial complexes and colored
# hypergraphs are supported.
ComplexType = CellComplex | CombinatorialComplex | ColoredHyperGraph


def node_diameters(domain: ComplexType) -> tuple[list[int], list[set[Hashable]]]:
    """Return the node diameters of the connected components in cell complex.

    Parameters
    ----------
    domain : CellComplex or CombinaorialComplex or ColoredHyperGraph
        The complex to be used to generate the node diameters for.

    Returns
    -------
    diameters : list
        List of the diameters of the s-components.
    components : list
        List of the s-component nodes.

    Examples
    --------
    >>> CC = tnx.CellComplex()
    >>> CC.add_cell([2, 3, 4], rank=2)
    >>> CC.add_cell([5, 6, 7], rank=2)
    >>> tnx.node_diameters(CC)
    >>> CCC = CC.to_combinatorial_complex()
    >>> tnx.node_diameters(CCC)
    >>> CHG = CC.to_colored_hypergraph()
    >>> tnx.node_diameters(CHG)
    """
    node_dict, A = domain.node_to_all_cell_adjacnecy_matrix(index=True)
    node_dict = {v: k for k, v in node_dict.items()}

    G = nx.from_scipy_sparse_array(A)
    diams = []
    comps = []
    for c in nx.connected_components(G):
        diamc = nx.diameter(G.subgraph(c))
        temp = set()
        for e in c:
            temp.add(node_dict[e])
        comps.append(temp)
        diams.append(diamc)
    return diams, comps


def cell_diameters(domain: ComplexType, s: int = 1) -> tuple[list[int], list[set[int]]]:
    """Return the cell diameters of the s_cell_connected component subgraphs.

    Parameters
    ----------
    domain : CellComplex or CombinatorialComplex or ColoredHyperGraph
        Supported complexes are cell/combintorial and hypegraphs.
    s : int, optional
        The number of intersections between pairwise consecutive cells.

    Returns
    -------
    list of diameters : list
        List of cell_diameters for s-cell component subcomplexes in the cell complex.

    list of component : list
        List of the cell uids in the s-cell component subcomplexes.

    Examples
    --------
    >>> CC = CellComplex()
    >>> CC.add_cell([2, 3, 4], rank=2)
    >>> CC.add_cell([5, 6, 7], rank=2)
    >>> tnx.cell_diameters(CC)
    >>> CCC = CC.to_combinatorial_complex()
    >>> tnx.cell_diameters(CCC)
    >>> CHG = CC.to_colored_hypergraph()
    >>> tnx.cell_diameters(CHG)
    """
    if not isinstance(domain, CellComplex | CombinatorialComplex | ColoredHyperGraph):
        raise TypeError(f"Input complex {domain} is not supported.")
    coldict, A = domain.all_cell_to_node_coadjacency_matrix(index=True)
    coldict = {v: k for k, v in coldict.items()}

    G = nx.from_scipy_sparse_array(A)
    diams = []
    comps = []
    for c in nx.connected_components(G):
        diamc = nx.diameter(G.subgraph(c))
        temp = set()
        for e in c:
            temp.add(coldict[e])
        comps.append(temp)
        diams.append(diamc)
    return diams, comps


def diameter(domain: ComplexType) -> int:
    """Return length of the longest shortest s-walk between nodes.

    Parameters
    ----------
    domain : CellComplex or CombinatorialComplex or ColoredHyperGraph
        Supported complexes are cell/combintorial and hypegraphs.

    Returns
    -------
    int
        The diameter of the longest shortest s-walk between nodes.

    Raises
    ------
    RuntimeError
        If the cell complex is not s-cell-connected

    Notes
    -----
    Two nodes are s-adjacent if they share s cells.
    Two nodes v_start and v_end are s-walk connected if there is a sequence of
    nodes v_start, v_1, v_2, ... v_n-1, v_end such that consecutive nodes
    are s-adjacent. If the cell complex is not connected, an error will be raised.

    Examples
    --------
    >>> CC = tnx.CellComplex()
    >>> CC.add_cell([2, 3, 4], rank=2)
    >>> CC.add_cell([5, 6, 7], rank=2)
    >>> CC.add_cell([2, 5], rank=2)
    >>> tnx.diameter(CC)
    >>> CCC = CC.to_combinatorial_complex()
    >>> tnx.diameter(CCC)
    >>> CHG = CC.to_colored_hypergraph()
    >>> tnx.diameter(CHG)
    """
    if not isinstance(domain, CellComplex | CombinatorialComplex | ColoredHyperGraph):
        raise TypeError(f"Input complex {domain} is not supported.")
    A = domain.node_to_all_cell_adjacnecy_matrix()
    G = nx.from_scipy_sparse_array(A)
    if nx.is_connected(G):
        return nx.diameter(G)
    raise RuntimeError("cc is not connected.")


def cell_diameter(domain: ComplexType, s: int = 1) -> int:
    """Return the length of the longest shortest s-walk between cells.

    Parameters
    ----------
    domain : CellComplex or CombinatorialComplex or ColoredHyperGraph
        Supported complexes are cell/combintorial and hypegraphs.
    s : int, default=1
        The number of intersections between pairwise consecutive cells.

    Returns
    -------
    int
        Returns the length of the longest shortest s-walk between cells.

    Raises
    ------
    RuntimeError
        If cell complex is not s-cell-connected

    Notes
    -----
    Two cells are s-coadjacent if they share s nodes.
    Two nodes e_start and e_end are s-walk connected if there is a sequence of
    cells (one or two dimensional) e_start, e_1, e_2, ... e_n-1, e_end such that consecutive cells
    are s-coadjacent. If the cell complex is not connected, an error will be raised.

    Examples
    --------
    >>> CC = tnx.CellComplex()
    >>> CC.add_cell([2, 3, 4], rank=2)
    >>> CC.add_cell([5, 6, 7], rank=2)
    >>> CC.add_cell([2, 5], rank=1)
    >>> tnx.cell_diameter(CC)
    >>> CCC = CC.to_combinatorial_complex()
    >>> tnx.cell_diameter(CCC)
    >>> CHG = CC.to_colored_hypergraph()
    >>> tnx.cell_diameter(CHG)
    """
    if not isinstance(domain, CellComplex | CombinatorialComplex | ColoredHyperGraph):
        raise TypeError(f"Input complex {domain} is not supported.")
    A = domain.all_cell_to_node_coadjacency_matrix()
    G = nx.from_scipy_sparse_array(A)
    if nx.is_connected(G):
        return nx.diameter(G)
    raise RuntimeError(f"cell complex is not s-connected. s={s}")



"""Test graph to simplicial complex transformation."""

import networkx as nx

from toponetx.transform.graph_to_simplicial_complex import (
    graph_to_clique_complex,
    graph_to_neighbor_complex,
    weighted_graph_to_vietoris_rips_complex,
)


class TestGraphToSimplicialComplex:
    """Test graph to simplicial complex transformation."""

    def test_graph_to_neighbor_complex(self):
        """Test graph_2_neighbor_complex."""
        G = nx.Graph()

        G.add_edge(0, 1)
        G.add_edge(1, 2)
        G.add_edge(2, 3)
        G.add_edge(3, 0)

        sc = graph_to_neighbor_complex(G)

        assert sc.dim == 2
        assert (0, 1) in sc
        assert (0, 2) in sc

    def test_graph_to_clique_complex(self):
        """Test graph_2_clique_complex."""
        G = nx.Graph()
        G.graph["label"] = 12

        G.add_node(0, label=5)
        G.add_edge(0, 1, weight=10)
        G.add_edge(1, 2)
        G.add_edge(2, 0)
        G.add_edge(2, 3)
        G.add_edge(3, 0)

        sc = graph_to_clique_complex(G)

        assert sc.dim == 2
        assert (0, 2, 3) in sc
        assert (0, 1, 2) in sc

        # attributes are copied
        assert sc.complex["label"] == 12
        assert sc[(0,)]["label"] == 5
        assert sc[(0, 1)]["weight"] == 10

        sc = graph_to_clique_complex(G, max_rank=1)

        assert sc.dim == 1
        assert (0, 2, 3) not in sc
        assert (0, 1, 2) not in sc

    def test_weighted_graph_to_vietoris_rips_complex(self):
        """Test weighted_graph_to_vietoris_rips_complex."""

        def generate_weighted_graph_for_vietoris_rips():
            """Create a weighted graph in networkx to test the Vietoris-Rips simplicial complex lift.

            Returns
            -------
            networkx graph
                A undirected weighted graph.
            """
            G = nx.Graph()
            # We add a 3-clique to the graph with pairwise weights 1.0
            G.add_edge(0, 1, weight=1.0)
            G.add_edge(0, 2, weight=1.0)
            G.add_edge(0, 3, weight=1.0)
            G.add_edge(1, 2, weight=1.0)
            G.add_edge(1, 3, weight=1.0)
            G.add_edge(2, 3, weight=1.0)
            # We add a 2-clique to the graph with pairwise weights 2.0
            G.add_edge(4, 5, weight=2.0)
            G.add_edge(4, 6, weight=2.0)
            G.add_edge(5, 6, weight=2.0)
            # We add an edge between the two cliques with weight 3.0
            G.add_edge(3, 4, weight=3.0)
            # We add a new vertex 7 to the graph that forms a 3 clique with the previous
            # vertices 4, 5, and 6 if we take r > 4.0
            G.add_edge(4, 7, weight=4.0)
            G.add_edge(5, 7, weight=4.0)
            G.add_edge(6, 7, weight=4.0)
            return G

        def generate_expected_simplices_for_vietoris_rips_complex(r):
            """Generate expected and unexpected simplices for the Vietoris-Rips complex of the test.

            This function returns a pair of lists of tuples containing the expected and unexpected simplices of the
            Vietoris-Rips persistence diagram of the graph generated with the function
            generate_weighted_graph_for_vietoris_rips depending on the radius r.

            Parameters
            ----------
            r : float
                Radius of the Vietoris-Rips complex.

            Returns
            -------
            expected_and_unexpected_simplices: tuple[list[tuple], list[tuple]]
                A tuple containing the lists of the expected and unexpected simplices (first and second lists,
                respectively) for the Vietoris-Rips complex.
            """
            expected_vertices = list(range(7))
            expected_edges = []
            unexpected_edges = []
            expected_triangles = []
            unexpected_triangles = []
            expected_tetrahedra = []
            unexpected_tetrahedra = []
            # Now, for each radius, we include or exclude simplices depending on its value.
            # Radius 1
            edges = expected_edges if r >= 1 else unexpected_edges
            triangles = expected_triangles if r >= 1 else unexpected_triangles
            tetrahedra = expected_tetrahedra if r >= 1 else unexpected_tetrahedra
            edges.extend([(0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3)])
            triangles.extend([(0, 1, 2), (0, 1, 3), (0, 2, 3), (1, 2, 3)])
            tetrahedra.append((0, 1, 2, 3))
            # Radius 2
            edges = expected_edges if r >= 2 else unexpected_edges
            triangles = expected_triangles if r >= 2 else unexpected_triangles
            edges.extend([(4, 5), (4, 6), (5, 6)])
            triangles.append((4, 5, 6))
            # Radius 3
            edges = expected_edges if r >= 3 else unexpected_edges
            edges.append((3, 4))
            # Radius 4
            edges = expected_edges if r >= 4 else unexpected_edges
            triangles = expected_triangles if r >= 4 else unexpected_triangles
            tetrahedra = expected_tetrahedra if r >= 4 else unexpected_tetrahedra
            edges.extend([(4, 7), (5, 7), (6, 7)])
            triangles.extend([(4, 5, 7), (4, 6, 7), (5, 6, 7)])
            tetrahedra.append((4, 5, 6, 7))
            expected_simplices = (
                expected_vertices
                + expected_edges
                + expected_triangles
                + expected_tetrahedra
            )
            unexpected_simplices = (
                unexpected_edges + unexpected_triangles + unexpected_tetrahedra
            )
            return expected_simplices, unexpected_simplices

        radii_to_check = [
            0,
            1,
            2,
            3,
            4,
        ]  # Four possible different configurations for the Vietoris-Rips complex of the
        # associated graph.
        weighted_graph = generate_weighted_graph_for_vietoris_rips()
        for radius in radii_to_check:
            sc = weighted_graph_to_vietoris_rips_complex(weighted_graph, radius)
            (
                expected_simplices,
                unexpected_simplices,
            ) = generate_expected_simplices_for_vietoris_rips_complex(radius)
            for simplex in expected_simplices:
                assert simplex in sc
            for simplex in unexpected_simplices:
                assert simplex not in sc



"""Base errors and exceptions for TopoNetX."""

__all__ = [
    "TopoNetXAlgorithmError",
    "TopoNetXException",
    "TopoNetXNoPath",
    "TopoNetXUnfeasible",
]


class TopoNetXException(Exception):
    """Base class for exceptions in TopoNetX."""


class TopoNetXAlgorithmError(TopoNetXException):
    """Exception that is raised if an algorithm terminates unexpectedly."""


class TopoNetXUnfeasible(TopoNetXAlgorithmError):
    """Exception raised by algorithms trying to solve a problem instance that has no feasible solution."""


class TopoNetXNoPath(TopoNetXUnfeasible):
    """Exception for algorithms that should return a path or path length where such a path does not exist."""
